/* file = degrade.c */ 
/* degrades some audio (.au) files,
   by mixing in some noise, read from another .au file */
/* Nigel Ward, University of Tokyo, March 1995 */

#include "cheap.h"
#include "util.h"
#include "auutil.h"

static int noise_len;
char raw_noise[50 * MAX_SAMPLES];
float noise[50 * MAX_SAMPLES];
float noise_level;
char target_dir[MAX_PATH_LEN];
int noise_cursor;        /* take successive chunks of noise from this file */
int same_name;

/* god alone knows why this declaration is needed */
double atof(char *str);

/* ----------------------------------------------------------------------------- */
int process_degrade_options(int argc, char *argv[]);
void degrade_usage(void);    
void main(int argc, char *argv[]);
void degrade_file(char *file); 

/* ----------------------------------------------------------------------------- */
int process_degrade_options(int argc, char *argv[])
{
  int i, last_arg_read;     

  same_name = FALSE;
  last_arg_read = 0;
  for (i = 1; i < argc && argv[i][0] == '-'; i++) {
    switch(argv[i][1]) {
    case 's':   same_name = TRUE;   break;
    case 'n':   i++; noise_level = (float) atof(argv[i]); last_arg_read++;   break;
    case 'd':   i++; strcpy(target_dir, argv[i]); last_arg_read++;   break;
    default:  
      fprintf(stderr, "   (%c is not a known option)\n", argv[i][1]);
      degrade_usage(); exit(STRANGE); break;
    } 
    last_arg_read++;  
  }
  verify_is_directory(target_dir);
  if (! (noise_level >= 0.0 && noise_level <= 1.0)) {
    fprintf(stderr, "degrade: %f is strange (expected a noise level between 0 and 1)\n",
	    noise_level);
    exit(STRANGE); }
  
  return(last_arg_read);
}

/*-----------------------------------------------------------------------------*/
void degrade_usage(void)    
{
  fprintf(stderr, "usage: \n");
  fprintf(stderr, "   degrade [options] noise_file [data_files]*\n");
  fprintf(stderr, "Options \n");
  fprintf(stderr, "  -d directory  specify output directory \n");
  fprintf(stderr, "  -n .xx        set fraction of output from noise (from .0 to 1.) \n");
  fprintf(stderr, "  -s            use output filenames same as input filenames (dangerous) \n");
  exit(STRANGE);
}

/* ----------------------------------------------------------------------------- */
void main(int argc, char *argv[])
{
  int start_arg, i;
  char ignore_info_string[MAX_INFO_STRING_LEN];

  start_arg = 1 + process_degrade_options(argc, argv);

  if(start_arg >= argc) 
    {fprintf(stderr, "degrade: missing noise file\n"); exit(STRANGE);}
  /* open noise file */
  noise_len = read_au_into_array(argv[start_arg], ignore_info_string, raw_noise, 50 * MAX_SAMPLES);
  convert_u2f(raw_noise, noise, 0, noise_len);
  start_arg++;
  
  /* the remaining arguments are a list of files to degrade */
  if (start_arg > argc) {
    fprintf(stderr, "match: no data files specified. exiting\n"); exit(STRANGE);}

  noise_cursor = 0;
  for(i = start_arg; i < argc; i++)
    degrade_file(argv[i]);

  exit(NORMAL);
}

/* ----------------------------------------------------------------------------- */
void degrade_file(char *file) 
{
  char info_string[MAX_INFO_STRING_LEN];
  char new_info_string[MAX_INFO_STRING_LEN];
  char raw_data[MAX_SAMPLES_MORE]; 
  float mixed_data[MAX_SAMPLES];
  char unsigned_mixed_data[MAX_SAMPLES];

  char new_filename[MAX_PATH_LEN], basename[MAX_PATH_LEN];
  char addendum[20];
  float data[MAX_SAMPLES_MORE];
  int input_len;
  int out_fd;
  int t;
  
  input_len = read_au_into_array(file, info_string, raw_data, MAX_SAMPLES);
  convert_u2f(raw_data, data, 0, input_len);

  for (t = 0; t < input_len; t++) {
    /* this ensures no overflow, but probably shrinks the volume */
      mixed_data[t] = 
	noise_level * noise[(noise_cursor + t) % noise_len]
	  + (1. - noise_level) * data[t]; }
  convert_f2u(mixed_data, unsigned_mixed_data, 0, input_len);
  noise_cursor += input_len;

  /* assemble output filename */
  strcpy(new_filename, target_dir);
  strcat(new_filename, "/");
  basefilename(basename, file);
  strcat(new_filename, basename);
  if (!same_name) {
    sprintf(addendum, "-d%d", (int) (noise_level * 100));
    strcat(new_filename, addendum); }
  strcat(new_filename, ".au");
  
  printf(" writing to file `%s'\n", new_filename);
  
  if (file_exists(new_filename)) {
    /* might be better _NOT_ to overwrite files */
    printf("  Warning: overwriting !!\n"); }
  
  out_fd = open_new_file(new_filename); 
  if (out_fd == -1) {
    perror("degrade: *** open failed ***");
    return; }
 
  strcpy(new_info_string, info_string);
  sprintf(addendum, "-degraded-%d%%", (int) (100 * noise_level));
  strcat(new_info_string, addendum);

  write_audio_hdr(out_fd, new_info_string, input_len);
  (void) write(out_fd, unsigned_mixed_data, input_len); 
  close(out_fd);
}

/* ----------------------------------------------------------------------------- */
