/* file = ept.c */
/* hypotheses where the endpoints of the signal are, given an .au file */ 
/* assumes that these is some silence at beginning and end of the file */
/* uses the stupidest, easy-to-implement algorithm I could think of */
/* Nigel Ward, University of Tokyo, May 1995 */

/* ----------------------------------------------------------------------------- */
#include "cheap.h"  
#include "util.h"
#include "auutil.h"  
#include "gmeat.h"  

#define MS_PER_CHUNK  20
#define CHUNK_SIZE    (SAMPLES_PER_MS * MS_PER_CHUNK)   
#define MAX_CHUNKS    ((int)(MAX_SAMPLES / CHUNK_SIZE))
#define THRESHOLD_RATIO .2

double atof(char *str);

void compute_start_end(char raw_data[], int nsamples);
void write_ws_file(char *au_filename, int signal_start, int signal_end, int end);
void write_ep_file(char *au_filename, int signal_start, int signal_end);

static char info_string[MAX_INFO_STRING_LEN];
static int start_ms, end_ms;

/* ----------------------------------------------------------------------------- */
static int fl_bigger (float *i, float *j)
{  return(*i > *j); }

/* ----------------------------------------------------------------------------- */
void ept_file(char *filename) 
{
  char raw_data[MAX_SAMPLES];

  int nsamples;

  nsamples = read_au_into_array(filename, info_string, raw_data, MAX_SAMPLES); 
  compute_start_end(raw_data, nsamples);

  printf("      start_ms %d, end_ms %d \n", start_ms, end_ms);
    
  /* printf("nsamples, %d, nms %d \n", nsamples, nsamples / SAMPLES_PER_MS); */
  write_ep_file(filename, start_ms, end_ms);
  write_ws_file(filename, start_ms, end_ms, nsamples / SAMPLES_PER_MS);
}

/* ----------------------------------------------------------------------------- */
void compute_start_end(char raw_data[], int nsamples)
{
  int chunk, nchunks;
  float signal_ness[MAX_CHUNKS];
  float sortable[MAX_CHUNKS];
  float typical_low, typical_high, threshold;

  nchunks = nsamples / CHUNK_SIZE;

  for (chunk = 0; chunk < nchunks; chunk++) { 
    signal_ness[chunk] = liveliness(raw_data, chunk * CHUNK_SIZE, CHUNK_SIZE);
    sortable[chunk] = signal_ness[chunk];
    /* printf("si[%2d] = %8.3f,  \n ", chunk, signal_ness[chunk]);   */
  }
  qsort(sortable, nchunks, sizeof(float), fl_bigger);
  /* for (chunk = 0; chunk < 4; chunk++) {
    printf("si[%d] = %f,   ", chunk, sortable[chunk]); } */


  typical_low = (sortable[0] + sortable[1] + sortable[2] + sortable[3]) / 4.0;
  typical_high = (sortable[nchunks - 1] + sortable[nchunks - 2] + 
		  sortable[nchunks - 3] + sortable[nchunks - 4]) / 4.0;
  /* printf("\n typical_low %f, typical_high = %f,   ", typical_low, typical_high);  */

  threshold = (1. - THRESHOLD_RATIO) * typical_low + THRESHOLD_RATIO * typical_high;
  /* printf("\n threshold = %f,   ", threshold);  */

  for (chunk = 0; chunk < nchunks; chunk++) {
    if (signal_ness[chunk] > threshold) {
      start_ms = chunk * MS_PER_CHUNK;
      break; } }
  
  for (chunk = nchunks -1 ; chunk > 0; chunk--) {
    if (signal_ness[chunk] > threshold) {
      end_ms = (chunk + 1) * MS_PER_CHUNK;
      break; }  }
}



/* ----------------------------------------------------------------------------- */
void write_ep_file(char *au_filename, int signal_start, int signal_end)
{
  FILE *out_fp = NULL;
  char out_path[MAX_PATH_LEN];

  make_new_path(out_path, au_filename, ".ep");
  out_fp = fopen_or_exit(out_path, "w"); 

  fprintf(out_fp, "endpoints %d %d\n", signal_start, signal_end);
}


/* ----------------------------------------------------------------------------- */
/* this function used to write results in a format allowing easy evaluation;
   important for tuning THRESHOLD_RATIO */

void write_ws_file(char *au_filename, int signal_start, int signal_end, int end)
{
/*  FILE *out_fp = NULL;
  char out_path[MAX_PATH_LEN];

  make_new_path(out_path, au_filename, ".ws");
  out_fp = fopen_or_exit(out_path, "w"); 

  write_wh_file_hdr(out_fp, info_string, end);
  write_wh_file_line(out_fp, "#", 0, signal_start - 0, .5);
  write_wh_file_line(out_fp, "#", signal_end, end - signal_end, .5);
*/
}

/* ----------------------------------------------------------------------------- */
void ept_usage(void)    
{
  fprintf(stderr, "usage: \n");
  fprintf(stderr, "   ept [input_file.au]*\n");
}


/* ----------------------------------------------------------------------------- */
int process_ept_args(int argc, char *argv[])
{
  int i;  int last_arg_read = 0;  

  for (i = 1; i < argc && argv[i][0] == '-'; i++) {
    switch(argv[i][1]) {
    default:      
      fprintf(stderr, "   (%c is not a known option)\n", argv[i][1]);
      ept_usage(); exit(STRANGE); break;
    } 
    last_arg_read =  i;
  }
  return(last_arg_read);
}

/* ----------------------------------------------------------------------------- */
void main(int argc, char *argv[])
{
  int i, start_arg;

  argv[argc] = NULL;
  start_arg = 1 + process_ept_args(argc, argv);

  if (start_arg >= argc) {
    fprintf(stderr,"feat: no .au files specified, exiting\n");
    exit(STRANGE);  } 
  for(i = start_arg; i < argc; i++) {
    printf( "  processing, path = %s \n", argv[i]);
    ept_file(argv[i]);
  }
  exit(NORMAL);
}
/*-----------------------------------------------------------------------------*/
