/* file = evec.c */ 
/* graphically shows a semantic interpretation file, and how close it matches
   a veridical semantic interpretation file */
/* Nigel Ward, University of Tokyo, Feb 1995 */

/*-----------------------------------------------------------------------------*/
#include "cheap.h" 
#include "util.h"
#include "xutil.h"
#include "auutil.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#define NUM_VERBS  4
#define MAX_CV_FILES 100 

#define FILENAME_YPOS   15
#define NOTICE_Y        FILENAME_YPOS
#define PF_LABEL_YPOS   (FILENAME_YPOS + 20)
#define FIRST_ROW_BASE  (PF_LABEL_YPOS + 22)

#define LEFT_MARGIN     20
#define NAME_STARTX     60
#define NOTICE_X       (LEFT_MARGIN + 200)
#define PROFILES_STARTX 120
#define SUM_STARTX      450

#define YSPACING         64
#define XSPACING         36
#define PRE_BOX_WIDTH    20
#define PRO_BOX_WIDTH    12
#define BOX_HEIGHT       50

#define CANVAS_WIDTH     560
#define CANVAS_HEIGHT    (FIRST_ROW_BASE + NUM_CONCEPTS * YSPACING + 55)

static int ybaseline;    /* current position to draw a concept at */

static int detailed;     /* show score for each concept */
static int this_file;
static int nfiles;
static char file_list[MAX_CV_FILES][MAX_PATH_LEN];
static float file_pre_score[MAX_CV_FILES];
static float file_pro_score[MAX_CV_FILES];
static float file_comb_score[MAX_CV_FILES];
static char veridical_dir[MAX_PATH_LEN];   /* directory in which to seek .vv files */
static int windowp = FALSE;
static int display_alphabetically = FALSE;

static int nconcepts;
static int nguesses;

char pf_names[NUM_PFEATURES][MAX_PF_LENGTH + 1] = 
  {"volitional", "active", "affected", 
    "responsible", "direct-cause", "partial-cause",
     "topic", "individuated", "recepient"};
char pf_abbrevs[NUM_PFEATURES][2 + 1] = 
   {"vo", "ac", "af", "rs", "dc", "pc", "tp", "in", "rc"} ;
char ignored_abbrevs[NUM_IGNORED_ABBREVS][2 + 1] = {"lg", "gg"};


char verbs[NUM_VERBS][MAX_WORD_LEN + 1] =
   {"ate", "attacked", "slept", "rescued"};

struct concept true_cv[CONCEPT_ARRAY_SIZE];
struct concept guess_cv[CONCEPT_ARRAY_SIZE];

static float scaled_pre_err, scaled_pro_err, comb_err;
float cum_pre_err;
float cum_pro_err;

/*-----------------------------------------------------------------------------*/
void main(int argc, char *argv[]);
int process_evec_options(int argc, char *argv[]);
void evec_usage(void);

void setup_for_new_file(void);
void prepare_to_compute_errs_for_file(void);

struct concept *correspondent(char *label, struct concept vec[], int vec_length);
float lookup_pf_score(struct concept *con_ptr, char *pf_abbrev);

static int presence_compare (struct concept *i, struct concept *j);
void loop_thru_files(void);
void loop_thru_concepts(void);

void do_exit_processing(void);
void compute_cumulatives(void);

void normalize_guessed_pre(void);
float compute_pre_err(struct concept *true_ptr, struct concept *guess_ptr);
float compute_pf_err(struct concept *true_ptr, struct concept *guess_ptr, int pf_num);
float compute_pro_err(struct concept *true_ptr, struct concept *guess_ptr); 
int is_verb(char *string);
float difference(float a, float b);

void check_pfeatures(struct concept cv[], int cnum);
int valid_abbrev(char *abbrev);

/* window stuff follows */
void event_loop(void);
void handle_command(char *string);
void print_help(void);

void prepare_canvas(void);
void clean_up_canvas(void);
void redraw_display(void);

void display_filename(void);
void draw_notice(char *string);
void erase_notice(void);
void open_next_file(void);
void open_previous_file(void);
void display_string_compact(int x, int y, char *string);

void display_cv(void);
void display_con(struct concept *true_ptr, struct concept *guess_ptr, char *name);
void display_presence(int ybaseline, struct concept *true_ptr, 
		      struct concept *guess_ptr, int scorep);
void display_pf(struct concept *true_ptr, struct concept *guess_ptr, 
		int pf_num, int current_x, int veridical_meaningfulp);
void display_cumulatives(void);


/*=============================================================================*/
void main(int argc, char *argv[])
{
  int i, first_cv_file;

  first_cv_file = 1 + process_evec_options(argc, argv);
  
  nfiles = argc - first_cv_file;
  if (nfiles == 0) {
    fprintf(stderr, "please specify one or more .cs files\n");
    evec_usage();
    exit(STRANGE);}
  if (nfiles == 1) 
    detailed = TRUE;     /* if only one file, show details of scoring */
  for (i = 0; i < nfiles; i++) {
    strcpy(file_list[i], argv[first_cv_file + i]);
    file_comb_score[i] = INVALID; }

  if (windowp) {
    prepare_canvas();
    this_file = 0;
    setup_for_new_file();
    event_loop(); }
  else {
    loop_thru_files();
    do_exit_processing(); }
}


/*-----------------------------------------------------------------------------*/
int process_evec_options(int argc, char *argv[])
{
  int i, last_arg_read;
  strcpy(veridical_dir, "");

  for (i = 1; i < argc && argv[i][0] == '-'; i++) {
    switch(argv[i][1]) {
    case 'a':  display_alphabetically = TRUE; break;
    case 'd':  case 'l': 
    case 'r':  strcpy(veridical_dir, argv[i+1]);  i++;  break;
    case 'w':  windowp = TRUE;  break; 
    default:  
      fprintf(stderr, "   (%c is not a known option)\n", argv[i][1]);
      evec_usage(); exit(STRANGE); break;
    }
    last_arg_read = i;
  }
  return(last_arg_read);
}

/*-----------------------------------------------------------------------------*/
void evec_usage(void)
{
  fprintf(stderr, "usage: \n");
  fprintf(stderr, "   evec [-w] [-d directory] {clue-file.cs}+\n");
  fprintf(stderr, "Options \n");
  fprintf(stderr, "  -w         window display\n");
  fprintf(stderr, "  -r         directory to look in for .vv files\n");
  fprintf(stderr, "  -a         display concepts alphabetically\n");
}

/*-----------------------------------------------------------------------------*/  
void do_exit_processing(void)
{
  int i, num_scored_files;
  float grand_pre, grand_pro, grand_comb;

  grand_pre = 0.0;  grand_pro = 0.0;  grand_comb = 0.0;
  num_scored_files = 0;
  for(i=0; i < nfiles; i++)
    if (file_comb_score[i] != INVALID) {
      grand_pre += file_pre_score[i];
      grand_pro += file_pro_score[i];
      grand_comb += file_comb_score[i];
      num_scored_files ++; } 
  printf("\nAverages over the %d scored files, pre %.3f, pro %.3f,  comb  %.3f\n",
	 num_scored_files, 
	 grand_pre / num_scored_files,
	 grand_pro / num_scored_files,
	 grand_comb / num_scored_files);
}

/*-----------------------------------------------------------------------------*/  
void loop_thru_files(void)
{
  for(this_file = 0; this_file < nfiles; this_file++) {
    if(detailed) {
      printf("\nFor %s: --------------------------------\n", file_list[this_file]);
      printf(  "   concept  scaled_guess_score  presence_error  profile_error\n"); }
    else
      printf("%30s:", file_list[this_file]);
    
    setup_for_new_file();
    if (nguesses != 0 && nconcepts != 0) {
      prepare_to_compute_errs_for_file();
      loop_thru_concepts();
      compute_cumulatives();
      if (detailed)
	printf("            cumulative errors: pre = %4.2f (raw=%4.2f), pro = %4.2f (raw=%4.2f)\n",
	       scaled_pre_err, cum_pre_err, scaled_pro_err, cum_pro_err);
      printf("  pre %5.2f,  pro %5.2f,   comb_e  %6.3f\n", 
	     scaled_pre_err, scaled_pro_err, comb_err);
    } } }
  

/*-----------------------------------------------------------------------------*/
void loop_thru_concepts(void)
{
  int i;  float pre_err, pro_err;
  struct concept *true_ptr, *guess_ptr; 

  for (i = 0; i < nconcepts; i++) {
    true_ptr = &(true_cv[i]);
    guess_ptr = correspondent(true_cv[i].name, guess_cv, nguesses);
    
    pre_err = compute_pre_err(true_ptr, guess_ptr);
    cum_pre_err += pre_err;
    
    if(!is_verb(true_ptr->name)) {
      pro_err = compute_pro_err(true_ptr, guess_ptr);
      cum_pro_err += pro_err;
      if (detailed)
	printf("   %-10s   %6.1f          %6.1f          %6.1f\n", 
	       true_ptr->name, guess_ptr->presence, pre_err, pro_err); }
    else {
      if (detailed)
	printf("   %-10s   %6.1f          %6.1f              -\n", 
	       true_ptr->name, guess_ptr->presence, pre_err); }
  } }

/*-----------------------------------------------------------------------------*/
/* return the concept in vec whose name matches the label given.
   if no such concept is found, return the zero-concept. */
struct concept *correspondent(char *label, struct concept vec[], int vec_length)
{
  int i;

  for (i = 0; i < vec_length; i++)
    if (strcmp(vec[i].name, label) == 0) {
      /* printf("found a match for `%s'\n",label); */ 
      return(&(vec[i])); }
  /* printf("returning null; no match for `%s'\n",label); */
  return(NULL);
}

/*-----------------------------------------------------------------------------*/
int is_verb(char *string)
{
  int i;

  for (i = 0; i < NUM_VERBS; i++)
    if(strcmp(string, verbs[i]) == 0)
      return(TRUE);
  return(FALSE);
}

/* ----------------------------------------------------------------------------- */
static int presence_compare (struct concept *i, struct concept *j)
{  return(i->presence < j->presence);  }

static int name_compare (struct concept *i, struct concept *j)
{  return(strcmp(i->name, j->name)); }
/* ----------------------------------------------------------------------------- */

void setup_for_new_file(void)
{  
  nguesses = read_cv_file("", file_list[this_file], guess_cv, ".cv",FALSE); 
  if (nguesses == 0)
    printf("\n   Warning: missing or strange guessed-concepts file %s \n", file_list[this_file]);
  check_pfeatures(guess_cv, nguesses);

  nconcepts = read_cv_file(veridical_dir, file_list[this_file], true_cv, ".vv", TRUE);
  if (nconcepts == 0)
    printf(" Warning: missing or strange .vv file for %s \n", file_list[this_file]);
  check_pfeatures(true_cv, nconcepts);
  
  /* sort to display higher-scored guesses first */
  if (display_alphabetically) {
    qsort(guess_cv, nguesses, sizeof(struct concept), name_compare); }
  else {
    qsort(guess_cv, nguesses, sizeof(struct concept), presence_compare); }

  normalize_guessed_pre();
}


/*-----------------------------------------------------------------------------*/
void normalize_guessed_pre(void)
{
  int i;
  float sum_guessed_pre;

  sum_guessed_pre = 0;
  for (i = 0; i < nguesses; i++)
      sum_guessed_pre += guess_cv[i].presence;
  if (nconcepts > 0)
    for (i = 0; i < nguesses; i++)
      guess_cv[i].presence = (nconcepts / sum_guessed_pre) * guess_cv[i].presence;
}

/*-----------------------------------------------------------------------------*/
void prepare_to_compute_errs_for_file(void)
{
  int i;
  cum_pre_err = 0.0;
  cum_pro_err = 0.0;

  for (i = 0; i < nconcepts; i++)
    true_cv[i].already_done = FALSE;
  for (i = 0; i < nguesses; i++)
    guess_cv[i].already_done = FALSE;
}

/*-----------------------------------------------------------------------------*/
float compute_pro_err(struct concept *true_ptr, struct concept *guess_ptr)
{
  int pf_num;   float pro_err;
  
  pro_err = 0.;
  for(pf_num = 0; pf_num < NUM_PFEATURES; pf_num++) {
    pro_err += compute_pf_err(true_ptr, guess_ptr, pf_num);}
  return(pro_err);
}

float compute_pf_err(struct concept *true_ptr, struct concept *guess_ptr, int pf_num)
{
  return (difference(lookup_pf_score(guess_ptr, pf_abbrevs[pf_num]), 
		     lookup_pf_score(true_ptr, pf_abbrevs[pf_num])));
}

float lookup_pf_score(struct concept *con_ptr, char *pf_abbrev)
{
  int i;

  for (i = 0; i < con_ptr->npfeatures; i++) {
    if (strcmp(con_ptr->pf_name[i],pf_abbrev) == 0) {
      return(con_ptr->pf_score[i]); } }
  /* not found */
  return(0.0);
}

/*-----------------------------------------------------------------------------*/
float compute_pre_err(struct concept *true_ptr, struct concept *guess_ptr)
{
  float true_pre = true_ptr->presence;
  float guess_pre = guess_ptr->presence;
  if (true_pre > 1.0001 || true_pre < .9999) {
    fprintf(stderr, "something weird %f", true_pre); exit(STRANGE);}
  if (guess_pre >= true_pre) return(0.0);
  else return(true_pre - guess_pre);
}
  
/* ----------------------------------------------------------------------------- */
float difference(float a, float b)
{
  return ((float) fabs( (double) a - b)); 
  /* return ((float) fabs( (double) (a - b) * (a - b)));  */
}

/* ----------------------------------------------------------------------------- */
void compute_cumulatives(void)
{
  int i, nparticipants;
  nparticipants = 0;
  for (i = 0; i < nconcepts; i++)
    if (!is_verb(true_cv[i].name))
      nparticipants ++;
  
  if (nparticipants == 0) nparticipants = 1;  /* prevent ridiculous scaling */
  scaled_pre_err = cum_pre_err / nconcepts;
  /* divide by 2 since the range here is from -1 to +1, so max err = 2 */
  scaled_pro_err = ((cum_pro_err / 2) / NUM_PFEATURES) / nparticipants;
  comb_err = (scaled_pre_err + scaled_pro_err) / 2.;
  file_pre_score[this_file] = scaled_pre_err; 
  file_pro_score[this_file] = scaled_pro_err; 
  file_comb_score[this_file] = comb_err; 
}


/*-----------------------------------------------------------------------------*/

void check_pfeatures(struct concept cv[], int cnum)
{
  int i, j;
  for (i = 0; i < cnum; i++)
    for (j = 0; j < cv[i].npfeatures; j++) {
      if (!valid_abbrev(cv[i].pf_name[j]))
	  fprintf(stderr, "Warning: %s is not a known feature (at `%s'); ignoring it\n", 
		  cv[i].pf_name[j], cv[i].name); }
}

int valid_abbrev(char *abbrev)
{
  int i;
 
  for (i = 0; i < NUM_PFEATURES; i++)
    if (strcmp(abbrev,pf_abbrevs[i]) == 0)
      return(TRUE);
  for (i = 0; i < NUM_IGNORED_ABBREVS; i++)
    if (strcmp(abbrev,ignored_abbrevs[i]) == 0)
      return(TRUE);
  return(FALSE);
}

/*=============================================================================
  below this point is the code for output to the window 
  =============================================================================*/
void event_loop(void)
{
  KeySym keysym;
  static char buffer[MAX_WORD_LEN];
  XComposeStatus compose;

  /* event loop code taken from Nye, volume 2 pg 72 */
  XSelectInput(display, win,
	       ExposureMask | KeyPressMask | ButtonPressMask | StructureNotifyMask);

  while (TRUE) {
    XNextEvent(display, &report);
    switch (report.type) {
    case Expose: 
      if (report.xexpose.count != 0)
	break;
      redraw_display();
      break;
    case KeyPress:
      (void) XLookupString( (XKeyEvent *) &report, buffer,
				MAX_WORD_LEN, &keysym, &compose);
      handle_command(XKeysymToString(keysym));
      break; 
    }
  }
}

/*-----------------------------------------------------------------------------*/
void handle_command(char *string)
{
  char cmd;   char msg_string[100];  

  erase_notice();    
  if (string == NULL) cmd = '?';

 if (strlen(string) == 1)     
    cmd = string[0];
  else {
    if (strcmp (string, "space") == 0) cmd = ' ';
    else if (strcmp (string, "semicolon") == 0) cmd = ';';
    else if (strcmp (string, "comma") == 0) cmd = ',';
    else if (strcmp (string, "BackSpace") == 0) cmd = '9';
    else if (strcmp (string, "Delete") == 0) cmd = '9';
    else {    
      sprintf(msg_string, "unknown command `%s': see term window for command list", string);
      draw_notice(msg_string);
      print_help();
      return ; } }

  switch (cmd) {
  case 'o': case 'n': case ' ': open_next_file(); break;
  case 'u': case 'p': case '9': open_previous_file(); break;
  case 'r': redraw_display(); break;
  case 'q': clean_up_canvas(); do_exit_processing(); exit(NORMAL); break;

  case '?': default:
    sprintf(msg_string, "unknown command `%c': see term window for command list", cmd);
    draw_notice(msg_string);
    print_help();
    break;
  }
}

/*-----------------------------------------------------------------------------*/
void print_help(void)
{
  printf("Commands:  \n");
  printf("     o - open next file (does w too) \n");
  printf("     u - reopen previous file \n");
  printf("     r - redraw display \n");
  printf("     q - quit (does w too) \n");
}


/*-----------------------------------------------------------------------------*/
void open_next_file(void)
{ if (this_file + 1 >= nfiles) 
    draw_notice("no more files; q to exit, then restart evec");
  else  {
    this_file += 1;
    setup_for_new_file();
    redraw_display();} }

/*-----------------------------------------------------------------------------*/
void open_previous_file(void)
{ if (this_file - 1 < 0)
    draw_notice("this is the first file, no previous ones");
  else {
    this_file -= 1;
    setup_for_new_file(); 
    redraw_display();} }


/*-----------------------------------------------------------------------------*/
int pf_yposition(int ybaseline, float score)
{
  return( (int) (ybaseline + (BOX_HEIGHT / 2) - (BOX_HEIGHT / 2) * score));
}

/*=============================================================================*/
/* Actual display routines start here */


/*-----------------------------------------------------------------------------*/
void display_cv(void)
{
  int i;

  prepare_to_compute_errs_for_file();

  display_filename();
  for (i = 0; i < NUM_PFEATURES; i++)
    display_string_compact(i * XSPACING + PROFILES_STARTX, PF_LABEL_YPOS, pf_names[i]);
  sg_string(SUM_STARTX, PF_LABEL_YPOS + FONT_HEIGHT, "sum_pro_errors");

  ybaseline = FIRST_ROW_BASE;
  if (display_alphabetically) { 
    for (i = 0; i < nguesses; i++) {
      display_con(correspondent(guess_cv[i].name, true_cv, nconcepts), 
		  &(guess_cv[i]),
		  guess_cv[i].name);} }

  else {    /* display in order */
    /* display correct guesses */
    for (i = 0; i < nconcepts; i++) {
      display_con(&(true_cv[i]), 
		  correspondent(true_cv[i].name, guess_cv, nguesses), 
		  true_cv[i].name);}
    /* display incorrect guesses (sorted by scores)*/ 
    ybaseline += 12;   /*leave gap to indicate start of incorrect guesses */
    for (i = 0; i < nguesses; i++) {
      if (!guess_cv[i].already_done)
	display_con(NULL, /* correspondent(guess_cv[i].name, true_cv, nconcepts), */
		    &(guess_cv[i]),
		    guess_cv[i].name);}
  }
  if (nconcepts > 0) {
    compute_cumulatives();
    display_cumulatives(); }
}


/*-----------------------------------------------------------------------------*/
void display_con(struct concept *true_ptr, struct concept *guess_ptr, char *name)
{
  int y_mid, pf_num, current_x, score_pro_p;
  float pro_err;
  char display_string[50];
  int scorep; 

  scorep = (int) true_ptr;

  y_mid = ybaseline + 3 + BOX_HEIGHT / 2;
  sg_string(NAME_STARTX, y_mid, name);

  display_presence(ybaseline, true_ptr, guess_ptr, scorep);

  score_pro_p = scorep && !is_verb(true_ptr->name);
  current_x = PROFILES_STARTX;
  for (pf_num = 0; pf_num < NUM_PFEATURES; pf_num++) {
    display_pf(true_ptr, guess_ptr, pf_num, current_x, score_pro_p);
    current_x += XSPACING; }

  if (score_pro_p) {
    pro_err = compute_pro_err(true_ptr, guess_ptr);
    sprintf(display_string, "%6.1f", pro_err);
    sg_string(SUM_STARTX, y_mid, display_string); 
    cum_pro_err +=pro_err; } 
  
  ybaseline += YSPACING;

  if (scorep) {
    cum_pre_err += compute_pre_err(true_ptr, guess_ptr);
    true_ptr->already_done = TRUE; }
  guess_ptr->already_done = TRUE;
  }


/*-----------------------------------------------------------------------------*/
int pre_yposition(int ybaseline, float score)
{
  return ((int) ybaseline + BOX_HEIGHT - BOX_HEIGHT * score); }     

    
void display_presence(int ybaseline, struct concept *true_ptr, 
		      struct concept *guess_ptr, int scorep)
{
  int y_zero, ytmp;

  y_zero = pre_yposition(ybaseline, 0);

  sg_outline_rect(LEFT_MARGIN, ybaseline, PRE_BOX_WIDTH, BOX_HEIGHT);
  
  /* indicate zero point */
  sg_line(LEFT_MARGIN - 2, y_zero, LEFT_MARGIN - 1, y_zero);
  sg_line(LEFT_MARGIN + PRE_BOX_WIDTH + 1, y_zero, LEFT_MARGIN + PRE_BOX_WIDTH + 2, y_zero);

  if (scorep) {
    ytmp = pre_yposition(ybaseline, true_ptr->presence);
    sg_draw_rect(LEFT_MARGIN + 2, ytmp - 1, PRE_BOX_WIDTH - 3, 3); }

  ytmp = pre_yposition(ybaseline, guess_ptr->presence);
  sg_line_red(LEFT_MARGIN, ytmp, LEFT_MARGIN + PRE_BOX_WIDTH, ytmp);
  sg_line_red(LEFT_MARGIN + 7, y_zero, LEFT_MARGIN + 7, ytmp);
  sg_line_red(LEFT_MARGIN + 13, y_zero, LEFT_MARGIN + 13, ytmp);
}


/*-----------------------------------------------------------------------------*/
void display_pf(struct concept *true_ptr, struct concept *guess_ptr, 
		int pf_num, int current_x, int veridical_meaningfulp)
{
  int ytmp, y_zero;
  float true_pro_score, guess_pro_score;

  /* write abbrev at bottom of box */
  sg_string(current_x, ybaseline + BOX_HEIGHT + FONT_HEIGHT + 2, pf_abbrevs[pf_num]);
  /* deemphasize irrelevant profiles by omitting box */
  if (veridical_meaningfulp)
    sg_outline_rect(current_x, ybaseline, PRO_BOX_WIDTH, BOX_HEIGHT);

  /* indicate zero point */
  y_zero = pf_yposition(ybaseline, 0);
  sg_line(current_x -2, y_zero, current_x - 1, y_zero);
  sg_line(current_x + PRO_BOX_WIDTH + 1, y_zero, current_x + PRO_BOX_WIDTH + 2, y_zero);

  if (veridical_meaningfulp) {
    true_pro_score = lookup_pf_score(true_ptr, pf_abbrevs[pf_num]);
    ytmp = pf_yposition(ybaseline, true_pro_score);
    sg_draw_rect(current_x + 2, ytmp - 1, PRO_BOX_WIDTH - 3, 3); }

  guess_pro_score = lookup_pf_score(guess_ptr, pf_abbrevs[pf_num]);
  ytmp = pf_yposition(ybaseline, guess_pro_score);
  sg_line_red(current_x, ytmp, current_x + PRO_BOX_WIDTH, ytmp);
  if (veridical_meaningfulp)
    sg_line_red(current_x + 6, pf_yposition(ybaseline, -1.), current_x + 6, ytmp);
}

/* ----------------------------------------------------------------------------- */
void display_cumulatives(void)
{
  char tmp_string[30];

  sg_string(LEFT_MARGIN, CANVAS_HEIGHT - 22, "cum_pre_error:"); 
  sprintf(tmp_string, "%.2f (raw = %.2f)", scaled_pre_err , cum_pre_err);
  sg_string(LEFT_MARGIN, CANVAS_HEIGHT - 10, tmp_string); 

  sg_string(SUM_STARTX, CANVAS_HEIGHT - 22, "cum_pro_error:"); 
  sprintf(tmp_string, "%.2f (raw = %.1f)", scaled_pro_err, cum_pro_err);
  sg_string(SUM_STARTX, CANVAS_HEIGHT - 10, tmp_string); 

  sg_string(CANVAS_WIDTH / 2 - 20, CANVAS_HEIGHT - 22, "> average <"); 
  sprintf(tmp_string, ">  %4.2f   <", comb_err);
  sg_string(CANVAS_WIDTH / 2 - 20, CANVAS_HEIGHT - 10, tmp_string); 
}


/*=============================================================================*/
/* here are some utilty-type window functions */
/*-----------------------------------------------------------------------------*/

void redraw_display(void)
{ sg_clear();
  display_cv(); }

/*-----------------------------------------------------------------------------*/
void display_filename(void)
{  char file_string[MAX_PATH_LEN + 10]; 
   sprintf(file_string, "file = %s", file_list[this_file]);
   sg_string(LEFT_MARGIN, FILENAME_YPOS, file_string); }

/*-----------------------------------------------------------------------------*/

/* this should use rotated text, but I'm to lazy to learn how to do that */
void display_string_compact(int x, int y, char *string)
{
  int i, yoffset;
  char tmp_string[6];
  tmp_string[5] = NULL;
  yoffset = y;
  for (i = 0; i < strlen(string); i += 5) {
    tmp_string[0] = string[i];
    tmp_string[1] = string[i+1];
    tmp_string[2] = string[i+2];
    tmp_string[3] = string[i+3];
    tmp_string[4] = string[i+4];
    sg_string(x, yoffset, tmp_string);
    yoffset += FONT_HEIGHT; }
}
    
/*-----------------------------------------------------------------------------*/
/* for warnings etc. */
void draw_notice(char *string)
{ erase_notice();
  sg_string_red(NOTICE_X, NOTICE_Y, string);}

void erase_notice(void)
{  sg_erase_rect(NOTICE_X, NOTICE_Y - FONT_HEIGHT, 1000, FONT_HEIGHT + 5); } 

void prepare_canvas(void)
{  sg_create(CANVAS_WIDTH, CANVAS_HEIGHT, 0, "evec"); }

void clean_up_canvas(void)
{ sg_close();}



/*=============================================================================*/
