;;; -*- Base: 10; Syntax: Common-lisp; Package: COMMON-LISP-USER; -*-

;;; file = file.lisp

;;; stuff for dealing with unix filesystem; 
;;; and especially for 
;;;   1. reading and writing word hypotheses (.wh/.gl), and  
;;;   2. reading and writing meaning vector (.gc/.gv/.vv/.tw) files

;;;=============================================================================
(defparameter *feature-file-ext* ".fe")

(defun is-au-file (pathname)
  (search ".au" (namestring pathname)))
(defun is-wh-file (pathname)
  (search ".wh" (namestring pathname)))
(defun is-gc-file (pathname)
  (search ".gc" (namestring pathname)))
(defun is-feature-file (pathname)
  (search *feature-file-ext* (namestring pathname)))
(defun is-label-file (pathname)
  (search ".la" (namestring pathname)))


;;;=============================================================================

;; same as the common lisp function pathname-name (except works only on unix)
;(defun last-component (path)
;  (let ((last-slash-pos (position #\/ path :from-end t)))
;    (if last-slash-pos
;	(subseq path (1+ last-slash-pos))
;      path) ) )

(defun file-exists (file)
  (with-open-file (stream file :direction :probe) t) )

;;; strip off extension, if any.
;;  there's probably a proper way to do this 
(defun strip-off-extension (filename)
  (let ((string (namestring filename)))
    (cond ((equal (aref string (- (length string) 3)) #\.)
	   (subseq string 0 (- (length string) 3)))
	  (t string))))

;;; example, given inputs "/home/sanpon...new.au" and  ".si", returns
;;; "/home/sanpon...new.si" 
(defun replace-extension (path new-extension)
  (concatenate 'string (strip-off-extension (namestring path)) new-extension))

;; read in data from a file,
(defun contents-of-file (path)
  (if (not (file-exists path))
      (error "in contents-of-file: file `~a' doesn't exist ~%" path)
    (with-open-file (stream path
			    :direction :input
			    :if-does-not-exist :error)
      (read stream nil) ) ) )

;;;=============================================================================
(defparameter *labels* nil)
(defun read-la-file (la-file)
  (setq *labels* nil)
  (with-open-file (stream la-file 
			  :direction :input
			  :if-does-not-exist :error)
    (let ((file-type-id (read-line stream)))
      (if (not (string-equal file-type-id *new-la-file-id-string*))
	  (error "~s not a la file (first line is ~s instead of ~s)~%"
		 la-file file-type-id *new-la-file-id-string*)))
    (parse-header stream nil) 
    ;; parse the labels
    (do ((exit-flag nil))
	(exit-flag)
	(multiple-value-bind (line eof) (read-line stream nil)  ; eof-error-p nil
			     ;;(format t "line is ~s ~%" line)
	     (cond ((or eof (null line))
		    ;;(format t "...eof detected ~%")
		    (setq exit-flag t))
		   (t (push (read-label-from-line line) *labels*)))))
    )
  (setq *labels* (nreverse *labels*))
  ) 

(defun read-label-from-line (line)
 (let ((sharp-macro-character (get-macro-character #\#)))
   (set-macro-character #\# (get-macro-character #\a))  ; make # temporarily normal
   (with-input-from-string (stream line)
     (let ((w-name (rename-silence (print-to-string (read stream))))
	   (w-start (ms-to-frames (read stream)))
	   (w-end (ms-to-frames (read stream))) )
       (set-macro-character #\# sharp-macro-character)
       (list  w-name w-start w-end)))))
  

;;;=============================================================================
;;; convert milliseconds to the nearest frame number
(defun ms-to-frames (milliseconds)
  (round (/ milliseconds *precision-reduction-factor*)))


(defun parse-header (stream use-headerp)
  (do ((exit-flag nil))
      (exit-flag)
      (let* ((input-line (read-line stream))
	     (colon-position (position #\: input-line))
	     (keyword (subseq input-line 0 colon-position))
	     (remainder (subseq input-line (1+ colon-position))) )
	;;(format t "keyword is ~s, remainder is ~s ~%" keyword remainder)
	(cond ((string-equal keyword "input_length")
	       (if use-headerp
		   (setq *input-len* (ms-to-frames (parse-integer remainder)))))
	      ((string-equal keyword "input_info_string")
	       (if use-headerp
		   (setq *input-info-string* 
			 (concatenate 'string remainder " (" *input-filebase* ".au" ")"))))
	      ((string-equal keyword "data")
	       (setq exit-flag t))
	      (t ;(format t "ignoring unknown keyword ~s ~%" keyword)
	       ))) ) )


;;;=============================================================================
(defun read-in-hypotheses (wh-file)
  (with-open-file (stream wh-file 
			  :direction :input
			  :if-does-not-exist :error)
    (let ((file-type-id (read-line stream)))
      (if (not (string-equal file-type-id *new-wh-file-id-string*))
	  (error "~s not a wh file (first line is ~s instead of ~s)~%"
		 wh-file file-type-id *new-wh-file-id-string*)))
    (parse-header stream t) 
    ;; parse the word hypotheses
    (setq *w-hyps* nil)
    (do ((exit-flag nil))
	(exit-flag)
	(multiple-value-bind (line eof) (read-line stream nil)  ; eof-error-p nil
			     ;;(format t "line is ~s ~%" line)
	     (cond ((or eof (null line))
		    ;;(format t "...eof detected ~%")
		    (setq exit-flag t))
		   (t (push (read-w-hyp-from-line line) *w-hyps*)))))
    )
  ;;(dolist (wh *w-hyps*) (format t "~%~s" wh) )

  (setq *num-w-hyps* (length *w-hyps*))
  (when (zerop *num-w-hyps*)
    (error "no hypotheses read in ") )
  (setq *w-hyps* (reverse *w-hyps*))
  ) 

(defun read-w-hyp-from-line (line)
 (let ((sharp-macro-character (get-macro-character #\#)))
   (set-macro-character #\# (get-macro-character #\a))  ; make # temporarily normal
   (with-input-from-string (stream line)
     (let ((w-name (print-to-string (read stream)))
	   (w-start (ms-to-frames (read stream)))
	   (w-end (ms-to-frames (read stream)))
	   (w-score (read stream)) )
       (set-macro-character #\# sharp-macro-character)
       (if (equal w-start w-end) (setq w-end (+ 1 w-end)))
       (let ((struct (make-whyp w-name w-start w-end w-score)))
	 (my-intern-and-set w-name struct)
	 struct)))))



(defun make-whyp (name start end score)
  (make-wh :name (rename-silence (strip-off-after-hyphen name))
	   :start start :length (- end start) :raw-score score
	   :garnet-objects (make-array 5 :initial-element nil)))  

(defun strip-off-after-hyphen (string)
  (let ((where (position #\- string)))
    (if where (subseq string 0  where)
      string)))

(defun rename-silence (string)
  (if (equal string "#") "sil" string) )

(defun unrename-silence (string)
  (if (equal string "sil") "#" string) )


;;;=============================================================================

(defun write-w-hyps (out-file score-access-fn)
    (with-open-file (stream out-file
			    :direction :output 
			    :if-exists :supersede
			    :if-does-not-exist :create)
      (format stream "~a~%" *new-wh-file-id-string*)
      (format stream "input_length: ~a~%"
	      (* *precision-reduction-factor* *input-len*))
      (format stream "input_info_string: from-file-~a ~%" *input-filebase*)
      (format stream "data:~%")
      (dolist (wh *w-hyps*)
	(format stream " ~a  ~d ~d ~f~%"
		(unrename-silence (string-downcase  (wh-name wh)))
		(* *precision-reduction-factor* (wh-start wh))
		(* *precision-reduction-factor* (+ (wh-start wh) (wh-length wh)))
		(funcall score-access-fn wh)))
      ) ) 

;;;=============================================================================

;;; write files in .gv format
(defun internal-to-file (out-file)
  (with-open-file (stream out-file               :direction :output 
			  :if-exists :supersede  :if-does-not-exist :create)
    (vec-to-c-style-file (internal-to-vec) stream *input-filebase*) ) )

(defun internal-to-file-stupid (out-file)
  (with-open-file (stream out-file               :direction :output 
			  :if-exists :supersede  :if-does-not-exist :create)
    (vec-to-c-style-file (internal-to-vec-stupid) stream *input-filebase*) ) )

(defun ordered-keys ()
  (append  ;(mapcar #'(lambda (r) (role-short-name r)) *roles*)
	  (mapcar #'(lambda (f) (feature-short-name f)) *pfeatures*) ) )

(defun vec-to-c-style-file (vec stream id-string)
  (format stream "~a~%" *cv-file-id-string*) 
  (format stream ";;input_info_string: from-file-~a ~%" id-string)
  (format stream ";;~%")
  (dolist (word-info vec)	
    (let ((word (first word-info)))
      (format stream "~(~a~) ~6,3f " word  (second word-info) ) 
      (when (is-noun (lookup-word word))
	(mapcar #'(lambda (key value)
		    (format stream " ~(~s~) ~6,3f " key value) )
		(ordered-keys)
		(cddr word-info) )
	) )
    (format stream "~%")
    )
  )

;;;=============================================================================

;;; reads a meaning-vector file, returns its contents in vec format
(defun file-to-vec (filename)
  (let ((concept-info nil))
    (with-open-file (stream filename
			    :direction :input
			    :if-does-not-exist :error)
    (let ((file-id (read-line stream)))
      (if (not (string-equal (subseq file-id 0 (length *cv-file-id-string*))
			     *cv-file-id-string*))
	  (error "~a is not a cv file ~%(first line is ~s)~%" filename file-id)))
    (do ((exit-flag nil))
	(exit-flag)
	(multiple-value-bind (line eof) (read-line stream nil)  ; eof-error-p nil
	     ;;(format t "line is ~s ~%" line)
	     (cond ((or eof (null line))
		    ;;(format t "...eof detected ~%")
		    (setq exit-flag t))
		   ((equal (length line) 0))    ; it's an empty line, so ignore it
		   ((equal (aref line 0) #\;))  ; it's a comment, so ignore it
		   (t (push (parse-vec-file-line line) concept-info)) ) ) )
    
    ;;(format t "concept-info ~s~%" concept-info)
    (dolist (word (all-concepts))
      (unless (assoc (word-name word) concept-info :test #'string-equal)
	      (push (cons (word-name word) 
			  (cons 0.0   ; default presence score
				(all-zero-profile)))
		    concept-info)))
    (setq concept-info (nreverse concept-info))
    concept-info
) ) )

(defun parse-vec-file-line (line)
  ;;(format t "parsing ~s,~% " line) 
  (let ((concept nil)
	(presence-score nil)
	(feat-value-pairs nil))
    (with-input-from-string (stream line)
      (multiple-value-bind (token eof) (read stream nil)  ; eof-error-p nil
			   (cond ((or eof (null line))) ; ignore blank lines
				 (t ;(format t "concept is ~s, " token)
				    (setq concept token))))
      (setq presence-score (read stream))
      (do ((exit-flag nil)) 
	  (exit-flag)
	  (multiple-value-bind (feature eof) (read stream nil)  ; eof-error-p nil
			       ;;(format t "feature is `~s'" feature)
			       (cond ((or eof (null feature))
				      (setq exit-flag t))
				     (t (push (list feature (read stream)) 
					      feat-value-pairs)
					))))
      )
    ;; (format t "~%feat-value-pairs ~s~%" feat-value-pairs)
    (cons (format nil "~(~s~)" concept)
	  (cons presence-score 
		(sort-fv-pairs feat-value-pairs)))
    ))

(defun sort-fv-pairs (pairs)
  (mapcar #'(lambda (long-short)
	      (let ((short (second long-short)))
		(if (assoc short pairs)
		    (second (assoc short pairs))
		  0.0)))
	  *feature-names*))

;;;=============================================================================
