/* file = labeler.c */
/* allows interactive labeling of speech data.
   Reads a .au file, generates a .la file */
/* Nigel Ward, University of Tokyo, April 1994 */

  /* contents:
     - data structures
     - .au file i/o
     - .la file i/o
     - basics of the display
     - processing user keystrokes
     - main */

#include "cheap.h"
#include "util.h"
#include "xutil.h"
#include "auutil.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
/*----------------------------------------------------------------------------- */
struct region *allocate_region(void);
struct region *make_and_link_in(int start, struct region *prev_reg_ptr,	char *label);

int current_region_start(void);
int current_region_end(void);
int next_region_start(struct region *reg_ptr);
int current_boundary(void);
void read_audio_file(void);
void play_active_region(void);
void play_visible_portion(void);
void play_two_sides(void);
void play_whole_input(void);
void write_labels(void);
void setup_bars(void);
void setup_labels(void);
void setup_default_regions(void);
void setup_regions_from_labels(int nlabels, struct whyp label_array[MAX_LABELS]);
void prepare_canvas(void);
void clean_up_canvas(void);
void redraw_display(void);
void display_filename(void);
void display_offset(void);
void display_signal(void);
void display_boundaries(void);
void display_labels(void);
int xpos_for_bar(int bar);
int visible(int bar);
void draw_bar(int x);
void do_special_boundary(int where, int draw);
void do_boundary(int where, int draw);
void clear_boundary(int where);
void draw_right_arrow(int x1, int x2, int y);    
void draw_left_arrow(int x1, int x2, int y);     
void decorate_boundary(int draw);         
void do_current_region_indicator(int draw);
void draw_notice(char *string);
void erase_notice(void);
void split_region(void);
void merge_regions(void);
void toggle_boundary(void);    
void move_active_boundary(int distance);  
void prompt_for_label_string(int draw);
void echo_input_string(void);
void draw_label(int where, char *string);
void edit_label(void);
void store_label(void);
void goto_next_region(void);
void goto_prev_region(void);
void open_next_file(void);
void open_previous_file(void);
void setup_for_new_file(void);
void handle_command(char *string);
void print_help(void);
void event_loop(void);
void main(int argc, char *argv[]);
void process_labeler_options(int argc, char *argv[]);
void labeler_usage(void);

/*----------------------------------------------------------------------------- */
/* wider displays are nicer but take longer to redisplay */
#define CANVAS_WIDTH       600  
#define Y_SCALING          .1
#define MAX_BOUNDARIES     (MAX_LABELS + 1)
#define MAX_REGIONS        (2 * MAX_LABELS + 1) /* allow silence betwen words */
#define DEFAULT_REGION_LEN 500  /* milliseconds */
#define SCROLLING_TRIGGER_THRESHOLD 40 /* pixels */
#define SCROLL_INCREMENT ((int) (CANVAS_WIDTH * .7))

/* define the origin point of the signal display */
#define HORIZ_MARGIN     20
#define NOTICE_X       (HORIZ_MARGIN + 200)
#define NOTICE_Y       40
#define FILENAME_YPOS            15
#define OFFSET_DISPLAY_YPOS      25
#define BOUNDARY_DECORATION_YPOS 50
#define INPUT_YPOS               65 
#define LABEL_YPOS               90
#define BOUNDARY_LINE_TOP        INPUT_YPOS 
#define LABEL_STRING_OFFSET      2

#define REGION_DECORATION_YPOS   93 
#define BOUNDARY_LINE_BOTTOM     410
#define MAX_NUMBER_BARS ((int) (MAX_SAMPLES_MORE / SAMPLES_PER_PIXEL))

#define SNIPPET_LEN           300      /* milliseconds */
#define PAUSE_LEN             300000   /* microseconds */ 

#define DRAW 1
#define ERASE 0
#define LEFT 0
#define RIGHT 1

/*----------------------------------------------------------------------------- */
int canvas_height;
int baseline;

/*-----------------------------------------------------------------------------*/
int old_style = FALSE;  /* boolean, describes how the audio data is displayed */

char info_string[MAX_INFO_STRING_LEN];
char raw_data[MAX_SAMPLES_MORE];
int all_bars[MAX_SAMPLES_MORE];
int nsamples, nbars;
int bar_height[MAX_NUMBER_BARS];
int read_in_progress;  /* boolean, describes current state of interface */

int active_boundary;   /* measured in pixels from start of data */
int offset;            /* current amount scrolled, in pixels */

int nfiles;
char file_list[100][MAX_PATH_LEN];
int this_file;

struct region {
  int start;
  char label[MAX_WORD_LEN];
  struct region *prev_region;
  struct region *next_region;
} ;

/* these regions are used to form  a double-linked list; 
   the reason why they are on an array is that I'm afraid of malloc */
struct region regions[MAX_REGIONS]; 
int free_pointer;
struct region *first_reg_ptr, *current_reg_ptr;

char string[MAX_WORD_LEN];       /* label, as input by user */

/*=============================================================================*/
struct region *allocate_region(void)
{
  struct region *reg_ptr;
  if (free_pointer >= MAX_REGIONS) 
    {fprintf(stderr, "!!Too many regions!! exiting ...\n"); exit(STRANGE);}
  reg_ptr = &(regions[free_pointer]);
  reg_ptr->start = 0; 
  reg_ptr->prev_region = NULL; 
  reg_ptr->next_region = NULL;
  reg_ptr->label[0] = NULL;
  free_pointer++;
  return(reg_ptr); }

/* creates a new region and links it into the list of regions */
struct region *make_and_link_in(int start, struct region *prev_reg_ptr, char *label)
{
  struct region *reg_ptr;

  reg_ptr = allocate_region();
  reg_ptr->start = start;
  strcpy(reg_ptr->label, label);
  if (start == 0) {
    first_reg_ptr = reg_ptr;}
  else {
    reg_ptr->prev_region = prev_reg_ptr;
    if (reg_ptr->prev_region != NULL) {
      (reg_ptr->prev_region)->next_region = reg_ptr;}}
  return(reg_ptr); }

/*-----------------------------------------------------------------------------*/
int current_region_start(void)
{  return(current_reg_ptr->start);}

int current_region_end(void)
{ if (current_reg_ptr->next_region == NULL) 
    return( nbars - 1);
  else
    return((current_reg_ptr->next_region)->start);}
     
int next_region_start(struct region *reg_ptr)
{  return(reg_ptr->next_region == NULL ? nbars : (reg_ptr->next_region)->start);}

int current_boundary(void)
{   return( (active_boundary == LEFT) ? current_region_start(): current_region_end());}


/*=============================================================================*/
void read_audio_file(void)
{
  char au_file_name[MAX_PATH_LEN];

  make_new_path(au_file_name, file_list[this_file], ".au");
  printf("opening file %s\n", au_file_name); 
  nsamples =  read_au_into_array(au_file_name, info_string, raw_data, MAX_SAMPLES_MORE);
  if (nsamples < 1) {perror("labeler: missing .au file"); exit(STRANGE); }
}

/*-----------------------------------------------------------------------------*/
int barnum_to_sample_num(int barnum)
{ return(barnum * SAMPLES_PER_PIXEL); }

/*-----------------------------------------------------------------------------*/
void play_active_region(void)
{
  int first_byte_at, last_byte_at;
  first_byte_at = barnum_to_sample_num(current_region_start());
  last_byte_at = barnum_to_sample_num(current_region_end());
  play_audio_data(raw_data, first_byte_at, last_byte_at - first_byte_at); }

void play_two_sides(void)
{
  int boundary_sample, num_samples;
  num_samples = SAMPLES_PER_MS * SNIPPET_LEN;
  boundary_sample = barnum_to_sample_num(current_boundary());

  play_audio_data(raw_data, boundary_sample - num_samples, num_samples);
  usleep(PAUSE_LEN);   /* .3 seconds */ 
  play_audio_data(raw_data, boundary_sample, num_samples);
}


void play_visible_portion(void)
{
  int start, end;

  if (offset == 0)     start = 0;
  else                 start = offset - HORIZ_MARGIN;
  if (nbars > CANVAS_WIDTH - HORIZ_MARGIN)     end = CANVAS_WIDTH - HORIZ_MARGIN;
  else                                         end = nbars;
  play_audio_data(raw_data, barnum_to_sample_num(start), barnum_to_sample_num(end)); } 

void play_whole_input(void)
{  play_audio_data(raw_data, 0, nsamples); }

void play_whole_input_with_pauses(void)
{
  struct region *reg_ptr;
  int start_sample, end_sample; 

  for (reg_ptr = first_reg_ptr; reg_ptr != NULL; reg_ptr = reg_ptr->next_region) {
    if (reg_ptr->label[0] != NULL) {
      start_sample = barnum_to_sample_num(reg_ptr->start); 
      if (reg_ptr->next_region == NULL) 
	end_sample = barnum_to_sample_num(nbars);
      else
	end_sample = barnum_to_sample_num((reg_ptr->next_region)->start);
      /* fprintf(stderr, "playing %d to %d\n", start_sample, end_sample); */
      play_audio_data(raw_data, start_sample, end_sample - start_sample);
      usleep(PAUSE_LEN); }
  } }


/*=============================================================================*/
void write_labels(void)
{
  char la_file_name[MAX_PATH_LEN];
  FILE *la_fp;
  struct region *reg_ptr;

  printf(" writing labels \n");
  make_new_path(la_file_name, file_list[this_file], ".la");
  la_fp = fopen_or_exit(la_file_name, "w");

  write_la_file_header(la_fp, info_string, nsamples / SAMPLES_PER_MS);
  for (reg_ptr = first_reg_ptr; reg_ptr != NULL; reg_ptr = reg_ptr->next_region) {
    if (reg_ptr->label[0] != NULL)
      write_la_entry(la_fp,
		     reg_ptr->label,
		     reg_ptr->start * DEFAULT_MS_PER_PIXEL, 
		     (reg_ptr->next_region == NULL)?
		     nsamples / SAMPLES_PER_MS  :
		     (reg_ptr->next_region)->start * DEFAULT_MS_PER_PIXEL); }
  fclose(la_fp);  }

/*-----------------------------------------------------------------------------*/
void setup_bars(void)
{
  float linearized[MAX_SAMPLES_MORE];
  int i, j;

  nbars = nsamples / SAMPLES_PER_PIXEL; 

  /* only needed for newstyle */
  for(i = 0; i < nsamples; i++) {
    all_bars[i] = (int) raw_data[i] * 1.;  
    if (all_bars[i] < 0) all_bars[i] = 128 - all_bars[i]; }

  /* only needed for oldstyle */
  convert_u2f(raw_data, linearized, 0, nsamples);
  for(j = 0; j < nbars; j++) {
    bar_height[j] = Y_SCALING * linearized[barnum_to_sample_num(j)];}
}

/*-----------------------------------------------------------------------------*/
void setup_labels(void)
{
  int nlabels; struct whyp label_array[MAX_LABELS]; int input_len;
  free_pointer = 0;
  input_len =  read_la_file(file_list[this_file], &nlabels, label_array, TRUE);
  if (input_len == 0) 
    setup_default_regions(); 
  else 
    setup_regions_from_labels(nlabels, label_array); 
}

/*-----------------------------------------------------------------------------*/
void setup_default_regions(void)
{
  int i;
  struct region *prev_reg_ptr;
  
  prev_reg_ptr = NULL;
  if (nbars == 0) {fprintf(stderr, "audio file too short\n"); exit(STRANGE);}
  for (i = 0; i < nbars; i += (int) DEFAULT_REGION_LEN / DEFAULT_MS_PER_PIXEL) {
    prev_reg_ptr = make_and_link_in(i, prev_reg_ptr, ""); } }


/*-----------------------------------------------------------------------------*/
void setup_regions_from_labels(int nlabels, struct whyp label_array[MAX_LABELS])
{ 
  int i, startpix, endpix, end_of_previous;
  struct region *prev_reg_ptr;

  end_of_previous = 0;   /* end of previous region */
  prev_reg_ptr = NULL;
  for (i = 0; i < nlabels; i++) {
    /* use this label as the description of a region */
    startpix = label_array[i].start / DEFAULT_MS_PER_PIXEL;
    endpix = label_array[i].end / DEFAULT_MS_PER_PIXEL;
    /* if this region does not abut the previous region, make an extra region */
    if (startpix != end_of_previous) {
      prev_reg_ptr = make_and_link_in(end_of_previous, prev_reg_ptr, "");
      end_of_previous = startpix; }
    prev_reg_ptr = make_and_link_in(startpix, prev_reg_ptr, label_array[i].label);
    end_of_previous = endpix;
  }
  if (end_of_previous  < nbars - 1 ) {
    (void) make_and_link_in(end_of_previous, prev_reg_ptr, ""); }
} 


/*=============================================================================*/
void prepare_canvas(void)
{ sg_create(CANVAS_WIDTH, canvas_height, 0, "labeler"); }

void clean_up_canvas(void)
{ sg_close();}

/*-----------------------------------------------------------------------------*/
void redraw_display(void)
{
  sg_clear();
  display_filename();
  display_offset();
  display_signal();
  display_labels();
  display_boundaries();
  do_current_region_indicator(DRAW);
  /* print_help(); */
  if (read_in_progress) {
    echo_input_string();
    prompt_for_label_string(DRAW); } }

/*-----------------------------------------------------------------------------*/
void display_filename(void)
{  char file_string[MAX_PATH_LEN + 10]; 
   sprintf(file_string, "file = %s", file_list[this_file]);
   sg_string(HORIZ_MARGIN, FILENAME_YPOS, file_string); }

void display_offset(void)
{  char o_string[100]; 
   if (offset != 0) {
     sprintf(o_string, "left edge of display is %d ms into the audio file ", offset);
     sg_string(HORIZ_MARGIN, OFFSET_DISPLAY_YPOS + FONT_HEIGHT + 2, o_string); }
 }
     
/*-----------------------------------------------------------------------------*/
void display_boundaries(void)
{
  struct region *reg_ptr;
  for (reg_ptr = first_reg_ptr; reg_ptr != NULL; reg_ptr = reg_ptr->next_region) {
    do_boundary(reg_ptr->start, DRAW); }
  do_boundary(nbars - 1, DRAW);
  decorate_boundary(DRAW); }

/*-----------------------------------------------------------------------------*/
void display_labels(void)
{
  struct region *reg_ptr;
  sg_erase_rect(0, LABEL_YPOS - FONT_HEIGHT, 1000, FONT_HEIGHT);

  for (reg_ptr = first_reg_ptr; reg_ptr != NULL; reg_ptr = reg_ptr->next_region) {
    if (reg_ptr->label[0] != NULL)
      draw_label(reg_ptr->start, reg_ptr->label);} }

void draw_label(int where, char *string)
{  sg_string(xpos_for_bar(where) + LABEL_STRING_OFFSET, LABEL_YPOS, string); }

/*-----------------------------------------------------------------------------*/
void display_signal(void)
{ int i;
  for (i = 0 ; i < nbars; i++) {
    if (visible(i))
      draw_bar(i); }
  sg_flush(); }


/*-----------------------------------------------------------------------------*/
int xpos_for_bar(int bar)
{ return(HORIZ_MARGIN - offset + bar);}

/* speed up things by avoiding drawing things that are outside of window */

int visible(int barnum)
{ return(((barnum - offset) > 0 - HORIZ_MARGIN) &&
	 ((barnum - offset) < CANVAS_WIDTH));}

void draw_bar(int bar)
{  
  int i;
  if (visible(bar)) {
    if (old_style)
      sg_line(xpos_for_bar(bar), baseline, xpos_for_bar(bar), baseline + bar_height[bar]); 
    else {
      for (i = 0; i < SAMPLES_PER_PIXEL; i++)
      sg_pset(xpos_for_bar(bar), baseline + all_bars[barnum_to_sample_num(bar) + i]); }
  }
}
/*-----------------------------------------------------------------------------*/
void do_special_boundary(int where, int draw)
{
  int x, y;
  x = xpos_for_bar(where);
  if (draw) {
    if (old_style) {
      for (y = BOUNDARY_DECORATION_YPOS - 2; y < BOUNDARY_LINE_BOTTOM; y += 8)
	sg_pset_red(x,y); 
      do_boundary(where, draw); }
    else sg_line(x, BOUNDARY_DECORATION_YPOS, x, BOUNDARY_LINE_BOTTOM); }
  else 
    clear_boundary(where);
}
  
void clear_boundary(int where)
{
  int x;
  x = xpos_for_bar(where);
  sg_clear_line(x, BOUNDARY_DECORATION_YPOS, x, BOUNDARY_LINE_BOTTOM);
  draw_bar(where);  
}

/*-----------------------------------------------------------------------------*/
/* if draw is true, then draw, else erase */
void do_boundary(int where, int draw)
{
  int x, y;
  if(visible(where)) {
    x = xpos_for_bar(where);
    /* draw a vertical bunch of dots */
    if(draw) {
      for (y = BOUNDARY_LINE_TOP; y < BOUNDARY_LINE_BOTTOM; y += 8)
	sg_pset(x,y); }
    else {
      clear_boundary(where);
    } } }

/*  ----------------------------------------------------------------------------- */
#define ARROW_LEN 10
#define ARROWHEAD_WIDTH 5

void draw_right_arrow(int x1, int x2, int y)       
{  sg_line(x1, y, x2, y);
   sg_line(x2 - ARROWHEAD_WIDTH, y - ARROWHEAD_WIDTH, x2, y);
   sg_line(x2 - ARROWHEAD_WIDTH, y + ARROWHEAD_WIDTH, x2, y); }

void draw_left_arrow(int x1, int x2, int y)     
{  sg_line(x1, y, x2, y);
   sg_line(x2 + ARROWHEAD_WIDTH, y - ARROWHEAD_WIDTH, x2, y);
   sg_line(x2 + ARROWHEAD_WIDTH, y + ARROWHEAD_WIDTH, x2, y); }

/*  ----------------------------------------------------------------------------- */
void decorate_boundary(int draw)         
{ 
  int x, y;
  x = xpos_for_bar(current_boundary());
  y = BOUNDARY_DECORATION_YPOS;
  if (draw) {
    draw_right_arrow(x, x + ARROW_LEN, y);
    draw_left_arrow(x, x - ARROW_LEN, y); 
    do_special_boundary(current_boundary(), DRAW); }
  else {
    sg_erase_rect(x - ARROW_LEN, y - ARROWHEAD_WIDTH,
		  2 * ARROW_LEN + 1, 2 * ARROWHEAD_WIDTH + 1);
    do_boundary(current_boundary(), ERASE); 
    do_boundary(current_boundary(), DRAW);  }
}

/*-----------------------------------------------------------------------------*/
/* for warnings etc. */
void draw_notice(char *string)
{ erase_notice();
  sg_string_red(NOTICE_X, NOTICE_Y, string);}

void erase_notice(void)
{  sg_erase_rect(NOTICE_X, NOTICE_Y - FONT_HEIGHT, 1000, FONT_HEIGHT + 5); } 

/*-----------------------------------------------------------------------------*/
void do_current_region_indicator(int draw)       
{
  int x1, x2;
  x1 = xpos_for_bar(current_region_start());
  x2 = xpos_for_bar(current_region_end());

  /* probably would look better stippled */
  if (draw)     sg_draw_rect(x1, REGION_DECORATION_YPOS, x2 - x1, 5);
  else         sg_erase_rect(x1, REGION_DECORATION_YPOS, x2 - x1, 5); }


/*=============================================================================*/
void split_region(void)
{
  int new_boundary;
  struct region *new_reg_ptr;

  if (current_region_end() - current_region_start() < 2) {
    sg_bell();
    draw_notice("region too small to split"); }
  else {
    decorate_boundary(ERASE);
    do_current_region_indicator(ERASE);
    /* splice in the new region */
    new_boundary = (current_region_end() + current_region_start()) / 2;
    do_boundary(new_boundary, DRAW);

    new_reg_ptr = allocate_region();
    new_reg_ptr->prev_region = current_reg_ptr;
    new_reg_ptr->next_region = current_reg_ptr->next_region;
    new_reg_ptr->start = new_boundary;
    if (current_reg_ptr->next_region != NULL) {
      (current_reg_ptr->next_region)->prev_region = new_reg_ptr; }
    current_reg_ptr->next_region = new_reg_ptr;
    do_current_region_indicator(DRAW);
    decorate_boundary(DRAW);
  } }

/*-----------------------------------------------------------------------------*/
void merge_regions(void)
{
  struct region *next_reg_ptr;
  do_current_region_indicator(ERASE);
  decorate_boundary(ERASE);            /* sometimes unnecessary */
  do_boundary(current_region_end(), ERASE); 
  next_reg_ptr = current_reg_ptr->next_region;
  if (next_reg_ptr == NULL) {
    draw_notice("sorry, no more regions"); }
  else {
    current_reg_ptr->next_region = next_reg_ptr->next_region;
    if (next_reg_ptr->next_region != NULL) {
      (next_reg_ptr->next_region)->prev_region = current_reg_ptr; }
  }
  do_current_region_indicator(DRAW);
  decorate_boundary(DRAW);
  display_labels();
}

/*-----------------------------------------------------------------------------*/
void toggle_boundary(void)    
{
  decorate_boundary(ERASE);

  if (active_boundary == LEFT && current_reg_ptr->next_region != NULL)
    active_boundary = RIGHT;
  else if (active_boundary == RIGHT && current_reg_ptr->prev_region != NULL)
    active_boundary = LEFT;
  else
    draw_notice("End boundaries not movable.  Do split perhaps");
  decorate_boundary(DRAW);
} 

/*-----------------------------------------------------------------------------*/
void move_active_boundary(int distance)  
{
  int tentative_new_pos;
  struct region *relevant_reg_ptr;

  if ((active_boundary == LEFT && current_reg_ptr->start == 0 )||
      (active_boundary == RIGHT && current_reg_ptr->next_region == NULL)) {
    sg_bell();
    draw_notice("End boundaries not movable.  Do a split perhaps"); }
  else {
    if (active_boundary == LEFT) relevant_reg_ptr = current_reg_ptr;
    else  relevant_reg_ptr = current_reg_ptr->next_region;
    tentative_new_pos = relevant_reg_ptr->start + distance; 
    if ((distance < 0 && tentative_new_pos <= (relevant_reg_ptr->prev_region)->start) ||
	(distance > 0  &&  tentative_new_pos >= next_region_start(relevant_reg_ptr))) {
      sg_bell();
      draw_notice("Cannot have zero length regions"); }
    else {
      decorate_boundary(ERASE);
      do_boundary(current_boundary(), ERASE);
      relevant_reg_ptr->start = tentative_new_pos;   
      do_boundary(current_boundary(), DRAW);
      decorate_boundary(DRAW);
      /* to save time, don't update active-region marker */
    } } }

/*-----------------------------------------------------------------------------*/
void prompt_for_label_string(int draw)
{
  int x1, width;
  x1 = current_region_start();
  width = current_region_end() - x1;

  if (draw) {
    draw_notice("enter label; return to end");
    sg_outline_rect(xpos_for_bar(x1), INPUT_YPOS - 5, width, FONT_HEIGHT + 5);  }
}

void erase_input_prompt(void)
{  sg_clear_outline_rect(xpos_for_bar(current_region_start()), INPUT_YPOS - 5,
			 current_region_end() -  current_region_start(), FONT_HEIGHT + 6);}

void echo_input_string(void)
{  
  sg_erase_rect(xpos_for_bar(current_region_start()), INPUT_YPOS - 1,
		xpos_for_bar(current_region_end()) , FONT_HEIGHT + 1);
  sg_string(xpos_for_bar(current_region_start()) + 2, INPUT_YPOS + 6, string);
}

/*-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------*/
void edit_label(void)
{  read_in_progress = TRUE;
  /* erase old label */
  sg_erase_rect(xpos_for_bar(current_region_start()), LABEL_YPOS - FONT_HEIGHT, 
		current_region_start() - current_region_end(), FONT_HEIGHT + 2);
  prompt_for_label_string(DRAW);}

/*-----------------------------------------------------------------------------*/
void store_label(void)
{
  read_in_progress = FALSE;
  strcpy(current_reg_ptr->label,string);
  string[0] = NULL;
  erase_input_prompt();
  /* draw_label(current_region_start(),current_reg_ptr->label); */
  display_labels();
  if (current_reg_ptr->next_region != NULL) {
    goto_next_region(); }      /* automatically go on to next region */
}

/*-----------------------------------------------------------------------------*/
void goto_next_region(void)
{
  if (current_reg_ptr->next_region == NULL) {
    draw_notice("no more regions"); }
  else {
    decorate_boundary(ERASE);
    do_current_region_indicator(ERASE);
    current_reg_ptr = current_reg_ptr->next_region;
    do_current_region_indicator(DRAW);
    active_boundary = LEFT;
    decorate_boundary(DRAW);   } }

void goto_prev_region(void)
{
  if (current_reg_ptr->prev_region == NULL)
    draw_notice("no more regions");
  else {
    decorate_boundary(ERASE);
    do_current_region_indicator(ERASE);
    current_reg_ptr = current_reg_ptr->prev_region;
    do_current_region_indicator(DRAW);
    active_boundary = RIGHT;
    decorate_boundary(DRAW);   } }

/*-----------------------------------------------------------------------------*/
void open_next_file(void)
{ if (this_file + 1 >= nfiles) 
    draw_notice("no more files; q to exit, then restart labeler");
  else  {
    write_labels();
    this_file += 1;
    setup_for_new_file(); } }

/*-----------------------------------------------------------------------------*/
void open_previous_file(void)
{ if (this_file - 1 < 0)
    draw_notice("this is the first file, no previous ones");
  else {
    write_labels();
    this_file -= 1;
    setup_for_new_file(); } }

/*-----------------------------------------------------------------------------*/
void setup_for_new_file(void)
{  read_audio_file();
   setup_bars();
   setup_labels();
   current_reg_ptr = first_reg_ptr;
   active_boundary = RIGHT;
   offset = 0;
   redraw_display(); }

/*=============================================================================*/
/* scroll if current region is too close to either edge of display */
void scroll_if_necessary(void)
{
  int should_move_right, should_move_left, new_offset;
  new_offset = offset;

  should_move_right = 
    (current_region_start() - offset < 0 + SCROLLING_TRIGGER_THRESHOLD
     && offset > 0); 
  should_move_left = 
    (current_region_end() - offset > CANVAS_WIDTH - SCROLLING_TRIGGER_THRESHOLD
     && offset < nbars - CANVAS_WIDTH);

  if (should_move_right && should_move_left) { /* do nothing */ }
  else if (should_move_right) {                new_offset = offset - SCROLL_INCREMENT; }
  else if (should_move_left) {                 new_offset = offset + SCROLL_INCREMENT; }

  if (new_offset != offset) {
    
    /* don't scroll data off screen */
    if (new_offset > (nbars - CANVAS_WIDTH + HORIZ_MARGIN)) 
      { new_offset = nbars - CANVAS_WIDTH + HORIZ_MARGIN * 2 ; }
    else if (new_offset < 0) 
      { new_offset = 0;}

    offset = new_offset;
    redraw_display(); }
}

/*=============================================================================*/
void handle_command(char *string)
{
  char cmd;   char msg_string[100];  

  erase_notice();    
  if (string == NULL) 
    cmd = '?';
  else if (strlen(string) == 1)     
    cmd = string[0];
  else {
    if (strcmp (string, "space") == 0) cmd = ' ';
    else if (strcmp (string, "semicolon") == 0) cmd = ';';
    else if (strcmp (string, "comma") == 0) cmd = ',';
    else if (strcmp (string, "BackSpace") == 0) cmd = '9';
    else if (strcmp (string, "Delete") == 0) cmd = '9';
    else {    
      sprintf(msg_string, "unknown command `%s': see term window for command list", string);
      draw_notice(msg_string);
      print_help();
      return ; } }

  switch (cmd) {
  case 'j': case 'K': move_active_boundary(-10); break;
  case 'k':           move_active_boundary(-2); break;
  case 'l':           move_active_boundary( 2); break;
  case ';': case 'L': move_active_boundary(10); break;
  case 't': toggle_boundary(); break; 

  case ' ': play_active_region();    break;
  case 'v': play_visible_portion();    break;
  case 'p': play_whole_input();    break;
  case 'c': play_whole_input_with_pauses();    break;
  case ',': play_two_sides();    break;
  case 'e': edit_label(); break;
  case 'f': goto_next_region(); break;
  case 'b': goto_prev_region(); break;
  case 's': split_region(); break;
  case 'm': merge_regions(); break;

  case 'w': write_labels(); break;
  case 'o': open_next_file(); break;
  case 'u': case '9': open_previous_file(); break;
  case 'r': redraw_display(); break;
  case 'q': write_labels(); clean_up_canvas(); exit(NORMAL); break;
  case 'A':                 clean_up_canvas(); exit(NORMAL); break;

  case '?': default:
    sprintf(msg_string, "unknown command `%c': see term window for command list", cmd);
    draw_notice(msg_string);
    print_help();
    break;
  }
}

/*-----------------------------------------------------------------------------*/
void print_help(void)
{
  printf("Commands:  \n");
  printf("  j ,K - move active boundary 10 to left      \n");
  printf("     k - move active boundary  2 to left \n");
  printf("     l - move active boundary  2 to right \n");
  printf("  ; ,L - move active boundary 10 to right \n");
  printf("     t - toggle active boundary (left or right of region) \n");
  printf(" \n");
  printf(" space - play active region (to speaker or jack) \n");
  printf("     v - play visible portion (to speaker or jack) \n");
  printf("     p - play entire file (to speaker or jack) \n");
  printf("     c - play all labeled regions in file (with pauses between) \n");  
  printf("     , - play %dms on either side of the active boundary \n", SNIPPET_LEN);  
  printf("     e - edit label for active region \n");
  printf("     f - move active region forward \n");
  printf("     b - move active region backward \n");
  printf("     s - split active region into two \n");
  printf("     m - merge active region with next region \n");
  printf(" \n");
  printf("     w - write labels \n");
  printf("     o - open next file (does w too) \n");
  printf("     u - reopen previous file \n");
  printf("     r - redraw display \n");
  printf("     q - quit (does w too) \n");
  printf("     A - abort \n"); 
}


/*-----------------------------------------------------------------------------*/
void event_loop(void)
{
  KeySym keysym;
  static char buffer[MAX_WORD_LEN];
  XComposeStatus compose;
  int length;

  /* event loop code taken from Nye, volume 2 pg 72 */
  XSelectInput(display, win,
	       ExposureMask | KeyPressMask | ButtonPressMask | StructureNotifyMask);

  while (TRUE) {
    XNextEvent(display, &report);
    switch (report.type) {
    case Expose: 
      if (report.xexpose.count != 0)
	break;
      redraw_display();
      break;
    case KeyPress:
      (void) XLookupString( (XKeyEvent *) &report, buffer,
				MAX_WORD_LEN, &keysym, &compose);
      /* printf ("ks=`%s'\n ", XKeysymToString(keysym)); */
      if (!read_in_progress) {
	/* it's a command */
	handle_command(XKeysymToString(keysym));
	scroll_if_necessary();
	break; }
      else { /* it's part of the label string */ /* code from Nye pg 270 */
	if ((keysym == XK_Return)    || (keysym  == XK_KP_Enter) ||
	    (keysym ==  XK_Linefeed) || (keysym == XK_Escape)) {
	  erase_notice();   /* remove the "input string" notice */
	  store_label();
	  break;}
	else if (((keysym >= XK_KP_Space) && (keysym <=  XK_KP_9)) ||
		 ((keysym >= XK_space) && (keysym <=  XK_asciitilde))) {
	  if ((strlen(string) + strlen(buffer)) >= MAX_WORD_LEN)
	    draw_notice("word label too long; subsequent characters ignored");
	  else {
	    strcat(string, buffer);
	    echo_input_string(); } }
	else if ((keysym == XK_BackSpace) || (keysym == XK_Delete)) {
	  if ((length = strlen(string)) > 0) {
	    string[length - 1] = NULL;
	    echo_input_string(); }
	  else sg_bell();}
	else if ((keysym == XK_Shift_L) || (keysym == XK_Shift_R)) {
	  /* ignore it */ }
	else {printf("Sorry, that key (%s) is not handled \n", XKeysymToString(keysym));}
	break;
      } } }
}

/*-----------------------------------------------------------------------------*/
void main(int argc, char *argv[]) 
{
  process_labeler_options(argc, argv);
  this_file = 0;
  print_help();
  prepare_canvas();
  setup_for_new_file();
  event_loop(); 
  /* not reached */
  exit(NORMAL); }

/*-----------------------------------------------------------------------------*/
void process_labeler_options(int argc, char *argv[]) { int i,
first_file_at;

  for (i = 1; i < argc && argv[i][0] == '-'; i++) {
    switch(argv[i][1]) {
    case 'o':  old_style = TRUE; break;
    default:  
      fprintf(stderr, "   (%c is not a known option)\n", argv[i][1]);
      labeler_usage(); exit(STRANGE); break;  }
  }
  if (old_style) {
    first_file_at = 2;
    canvas_height = 700;
    baseline = 250; }
  else {
    first_file_at = 1;
    canvas_height = 400;
    baseline = 100; }
  nfiles = argc - first_file_at;
  if (nfiles == 0) {
    fprintf(stderr, "please specify one or more .au files\n");
    exit(STRANGE); }
  for (i = 0; i < nfiles; i++) {
    strcpy(file_list[i], argv[i+first_file_at]);
    printf("file_list[%d] = %s \n", i, file_list[i]);   }
}

/*-----------------------------------------------------------------------------*/
void labeler_usage(void)
{
  fprintf(stderr, "usage: \n");
  fprintf(stderr, "   labeler [-o] {audio-file.au}+\n");
  fprintf(stderr, "Options \n");
  fprintf(stderr, "  -o         old-style display of audio data\n");
}
/*-----------------------------------------------------------------------------*/
