/* file = mmeat.c --- the meat of the match (word spotting) operation */
/* Nigel Ward, University of Tokyo, April 1994 */

#include "cheap.h"
#include "util.h"
#include "xutil.h"
#include "fmeat.h"
#include "mmeat.h"
#include "rwin.h"
#include <math.h> 

/*-----------------------------------------------------------------------------*/

int setup_template(char *fullpath);
void complain_and_exit(struct signature *s_ptr);
float cumulative_energy(float energy_array[], int start, int nframes);
void match_at_time(struct signature *i_ptr, struct signature *t_ptr, int offset);
void show_best_hypothesis_per_template(int ordering[]);
static int name_compare (int *i, int *j);
void alphabetize_templates(void); 
void sort_templates_by_scores(void); 
void write_hypotheses_to_file (struct signature *s_ptr, FILE *file_ptr);
void invert_scores(void);
void correct_for_silence(void);
void emphasize(void);
int is_silence(char *info_string);
void make_template(char *info_string, int start_frame, int nframes, float score);
int detect_first_non_silence_frame(float threshold, float sum[], int nframes);
int detect_last_non_silence_frame(float threshold, float sum[], int nframes);
static int fl_bigger (float *i, float *j);

/*-----------------------------------------------------------------------------*/


/* ========= some stuff used by both real and match ========== */

#define LARGE_FLOAT 9999999999999999999999.0;       
#define NOT_SET -1

/* the larger this, the more preference for longer words */
#define LENGTH_FACTOR 1.8    /* was 2.0 */   /* good value determined empirically */
/* factor for going from raw-ish match scores to scores for output */
#define EMPHASIS_FACTOR 9.0     /* good value determined empirically */
/* scores for silence hypotheses tend to be inordinately high; here we correct */ 
#define SILENCE_FACTOR 4        /* good value determined empirically */

static struct signature templates[MAX_NTEMPLATES];
static int ntemplates;
static int most_likely_offset;        /* saved across templates */

static int numerical_index[MAX_NTEMPLATES];
static int alphabetical_index[MAX_NTEMPLATES];

static int nwh;    /* number of word hypotheses */

/*-----------------------------------------------------------------------------*/
int process_template_filenames(int argc, char *argv[], int start_arg)    /* ENTRY POINT */
{
  if(verbosep) fprintf(stderr, " match: first template file is argv[%d]=%s \n",
		       start_arg, argv[start_arg]); 
  if (start_arg == argc) {
    fprintf(stderr, "match: no templates specified. exiting\n"); exit(STRANGE);}
  else {
    ntemplates = 0;
    while (start_arg < argc) {
      (void) setup_template(argv[start_arg]);
      start_arg++; } }
  if (verbosep) fprintf(stderr, " read %d templates\n", ntemplates);
  return(templates[0].logp);
}

/* ----------------------------------------------------------------------------- */
int setup_template(char *fullpath)
{
  int result;   FILE *fp;

  if (verbosep) fprintf(stderr,"  match: reading %s \n", fullpath);

  if (ntemplates >= MAX_NTEMPLATES) {
    fprintf(stderr, "match: template limit number (%d) exceeded \n", MAX_NTEMPLATES);
    fprintf(stderr, "       ignoring subsequent templates \n");
    return(0);}

  fp = fopen(fullpath, "r");
  if (fp == NULL) {
    fprintf(stderr,"match: can't open template file '%s'\n", fullpath );
    perror("match");
    return(FALSE);}

  result = read_fe_into_sig(fp, &(templates[ntemplates]));
  fclose(fp);
  if (result == FALSE) {
    fprintf(stderr,"match: something wrong was detected reading %s\n", fullpath);
    return(FALSE); }

  if (ntemplates > 0) check_compatibility(&(templates[ntemplates]));
  
  /* the idea behind length_factor is 
     1. normalize the distance w.r.t length, (divide badness by length)
     2. reward matches for longer templates (divide badness by length again) */
  templates[ntemplates].length_factor =
    1.0 / pow((float)templates[ntemplates].nframes, LENGTH_FACTOR);
  ntemplates++;
  return(TRUE); 
}

/* ----------------------------------------------------------------------------- */
/* check that a signature is in the same format as templates[0]
   ie, verify that the features were computed in the same way */
void check_compatibility(struct signature *s_ptr)                  /* ENTRY POINT */
{  
  if (strcmp(templates[0].feature_type, s_ptr->feature_type) != 0) {
    fprintf(stderr,"  features computed incompatibly: feature_type %s != %s\n",
	    templates[0].feature_type, s_ptr->feature_type);
    complain_and_exit(s_ptr);}
  if (templates[0].nbins != s_ptr->nbins) {
    fprintf(stderr,"  features computed incompatibly: nbins %d != %d\n",
	    templates[0].nbins, s_ptr->nbins); 
    complain_and_exit(s_ptr);}
  if (templates[0].frame_period != s_ptr->frame_period) {
    fprintf(stderr,"  features computed incompatibly: frame_period %d != %d\n",
	    templates[0].frame_period, s_ptr->frame_period); 
    complain_and_exit(s_ptr);}
  if (templates[0].logp != s_ptr->logp) {
    fprintf(stderr,"  features computed incompatibly: with/without taking log\n"); 
    complain_and_exit(s_ptr);}
  if (templates[0].samples_per_ms != s_ptr->samples_per_ms) {
    fprintf(stderr,"  features computed incompatibly: samples per ms %d != %d\n",
	    templates[0].samples_per_ms, s_ptr->samples_per_ms);
    complain_and_exit(s_ptr);}
  if (templates[0].samples_per_frame != s_ptr->samples_per_frame) {
    fprintf(stderr,"  features computed incompatibly: samples per frame %d != %d\n",
	    templates[0].samples_per_frame, s_ptr->samples_per_frame);
    complain_and_exit(s_ptr);}
}

void complain_and_exit(struct signature *s_ptr)
{
  fprintf(stderr,"  so matching template %s is senseless\n", s_ptr->info_string);
  exit(STRANGE); }

/* ----------------------------------------------------------------------------- */
void pre_match_initialization(void)    /* ENTRY POINT */
{ int tnum;

  for (tnum = 0; tnum < ntemplates; tnum++) {
    templates[tnum].cutoff = LARGE_FLOAT;
    templates[tnum].best_score = LARGE_FLOAT;
    templates[tnum].best_offset = NOT_SET;  }
    nwh = ntemplates; }

/* ----------------------------------------------------------------------------- */
void incremental_match(int start, int end, struct signature *input_ptr)     /* ENTRY POINT */
     /* start/end are the start/end of the newly computed frames */
{
  int start_idx, end_idx, tnum;     
  struct signature *t_ptr;
  if (verbosep) fprintf(stderr, " mmeat: start %d, end %d\n", start, end);

  /* !! probable off-by-one errors */
  for (end_idx = start; end_idx <= end; end_idx++) {
    /* try all placements where the template's right edge
       falls within the newly computed frames */
    for (tnum = 0; tnum < ntemplates; tnum++) {
      t_ptr = &(templates[tnum]);
      start_idx = end_idx - t_ptr->nframes + 1;
      if (0 <= start_idx && end_idx <= end) {
	match_at_time(input_ptr, t_ptr, start_idx);
	if (verbosep)
	   fprintf(stderr,"start_idx %d, tnum %d , t_ptr->hyp_score[start_idx] %f\n",
		   start_idx, tnum, t_ptr->hyp_score[start_idx]);  
	if (realtime_winp)
	  draw_likelihood(start_idx, tnum, t_ptr->hyp_score[start_idx], t_ptr->logp); }
    }
    if (realtime_winp) sg_flush();
  }
}

/* ----------------------------------------------------------------------------- */
float cumulative_energy(float energy_array[], int start, int nframes)
{
  float sum; int i;
  sum = 0.0;
  for (i = 0; i < nframes; i++)
    sum += energy_array[start + i];
  return(sum); }

/* ----------------------------------------------------------------------------- */
void match_at_time(struct signature *i_ptr, struct signature *t_ptr, int offset)
{
  /* distance = raw difference; score = distance normalized */
  float input_val, templ_val, raw, distance, score, energy_ratio;  
  int time, bin, hope, using_log;
  using_log = i_ptr->logp;

  if (normalizep) {
    if (using_log)
      energy_ratio =
	(cumulative_energy(i_ptr->energy, offset, t_ptr->nframes) 
	 - cumulative_energy(t_ptr->energy, 0, t_ptr->nframes))
	  / t_ptr->nframes;
    else
      energy_ratio =
	cumulative_energy(i_ptr->energy, offset, t_ptr->nframes) 
	  / cumulative_energy(t_ptr->energy, 0, t_ptr->nframes); }
  else {energy_ratio = 1.0;}
  /* fprintf(stderr, "energy ratio is %.2f\n", energy_ratio); */
  distance = 0.0;
  hope = TRUE;
  for (time = 0; time < t_ptr->nframes && hope; time++) {
    for (bin = 0; bin < i_ptr->nbins; bin++) {
      input_val = i_ptr->bins[offset + time][bin];
      if (using_log)
	templ_val = energy_ratio + t_ptr->bins[time][bin];
      else
	templ_val = energy_ratio * t_ptr->bins[time][bin];
      raw = input_val - templ_val;
      distance += raw * raw; 
    } 
    if (prunep && (distance > t_ptr->cutoff) ) {
      /* the cumulative distance computed so far exceeds the best_score, 
	 meaning that this placement cannot be the best for this template, 
	 so give up on this placement (ie, prune search) now */
      /* fprintf(stderr, "pruned at %2d into %3d", time, offset); */
      hope = FALSE; }
  }
  score = distance * t_ptr->length_factor; 
  /* fprintf (stderr, "   tp: %10s, offset %3d, score %10.1f, best %10.1f, pruned %d \n",
	   t_ptr->info_string, offset, score, t_ptr->best_score, !hope);    */

  t_ptr->hyp_score[offset] = score;
  if (score < t_ptr->best_score) {
    t_ptr->cutoff = distance;
    t_ptr->best_score = score;
    t_ptr->best_offset = offset;
    most_likely_offset = offset; }
}


/*-----------------------------------------------------------------------------*/
void sort_and_output_templates(struct signature *s_ptr, FILE *output_fp)      /* ENTRY POINT */
{
  correct_for_silence();
  if (!old_style) invert_scores();
  if (!old_style) emphasize();

  if (verbosep) {
    alphabetize_templates();
    printf("\n Best matching locations for each template (alphabetical order): \n");
    show_best_hypothesis_per_template(alphabetical_index); }

  sort_templates_by_scores();
  if (verbosep) {
    printf("\n Best matching locations for each template (ranked): \n");
    show_best_hypothesis_per_template(numerical_index); }
  if (verbosep) fprintf(stderr, " match: ntemplates %d \n",ntemplates);

  write_hypotheses_to_file(s_ptr, output_fp); 
}


/* ----------------------------------------------------------------------------- */
/* this adjusts scores for silence hypotheses (which arise if the templates 
   include templates for silence, i.e., those for #) */
void correct_for_silence(void)
{
  int i;
  for (i = 0; i < ntemplates; i++) 
    if (is_silence(templates[i].info_string))
      templates[i].best_score = SILENCE_FACTOR * templates[i].best_score;
}

int is_silence(char *info_string)
{
  char label[MAX_WORD_LEN];
  clip_word_name(label, info_string);
  return (strcmp(label, "#") == 0);
}

/* ----------------------------------------------------------------------------- */
/* we invert the sense, so that better matches have higher scores,
   and scale into the range from 0 to 1 */
void invert_scores(void)
{
  int i;
  float min;

  min = templates[0].best_score;
  for (i = 0; i < ntemplates; i++) 
    if(templates[i].best_score < min)
      min = templates[i].best_score;
      
  for (i = 0; i < ntemplates; i++) 
    templates[i].best_score = min / templates[i].best_score;
}

/* ----------------------------------------------------------------------------- */
void show_best_hypothesis_per_template(int ordering[])
{
  int i;   struct signature templ;

  for(i = 0; i < ntemplates; i++) {
    templ = templates[ordering[i]];
    if (templ.best_offset != NOT_SET) {
      printf("   least distance (%13.6f) at %3d, for `%s'\n",
	     templ.best_score, templ.best_offset, templ.info_string); }
  } }  

/* ----------------------------------------------------------------------------- */
static int name_compare (int *i, int *j)
{  return(strcmp(templates[*i].info_string, templates[*j].info_string));  }

/* ----------------------------------------------------------------------------- */
void alphabetize_templates(void) 
{
  int i;
  for (i = 0; i < ntemplates; i++) {
    alphabetical_index[i] = i; }
  qsort(alphabetical_index, ntemplates, sizeof(int), name_compare); }
      
/* ----------------------------------------------------------------------------- */
static int num_smaller_better (int *i, int *j)
{  return(templates[*i].best_score > templates[*j].best_score);  }

static int num_larger_better (int *i, int *j)
{  return(templates[*i].best_score < templates[*j].best_score);  }

/* ----------------------------------------------------------------------------- */
static int fl_bigger (float *i, float *j)
{  return(*i > *j); }


/* ----------------------------------------------------------------------------- */
void sort_templates_by_scores(void) 
{
  int i;
  for (i = 0; i < nwh; i++) {
    numerical_index[i] = i; }
  if (old_style)
    qsort(numerical_index, nwh, sizeof(int), num_smaller_better); 
  else
    qsort(numerical_index, nwh, sizeof(int), num_larger_better); 
}

/* ----------------------------------------------------------------------------- */
void emphasize(void)
{
  int i;
  for (i = 0; i < ntemplates; i++) {
    /* these functions give equivalent results */
    /* templates[i].best_score = pow(templates[i].best_score, EMPHASIS_FACTOR);  */
      templates[i].best_score = pow(500000, (templates[i].best_score - 1.0) ); 
  }  
}

/* ----------------------------------------------------------------------------- */
void write_hypotheses_to_file (struct signature *s_ptr, FILE *file_ptr)
{
  int ctr, i;  
  
  write_wh_file_hdr(file_ptr, s_ptr->info_string,
		    s_ptr->nsamples / s_ptr->samples_per_ms);
  for (ctr = 0; ctr < nwh; ctr++) {
    i = numerical_index[ctr];
    if (templates[i].best_offset != NOT_SET) 
      write_wh_file_line(file_ptr, 
			 templates[i].info_string,
			 templates[i].best_offset * s_ptr->frame_period
			  + (s_ptr->samples_per_frame / s_ptr->samples_per_ms) / 2.0, 
			 templates[i].nframes * s_ptr->frame_period,
			 templates[i].best_score); }
  fclose(file_ptr);
}


/* ============================================================================= */
/* ========= some stuff called only from match.c ========== */

/* Most_likely_offset is used to allow setting a good value for
   best_score, in order to make pruning more effective.  Of course, this
   is only useful if the previous template was for the same word (eg,
   missed-from-jmm and missed-from-mmj).  If filenames are chosen sensibly 
   (so that, when alphabetized, templates for the same word are adjacent) 
   this will work.  The observed speedup due to this is about 10% */
void match_by_templates(struct signature *input_ptr)
{
  int index, tnum;
  struct signature *t_ptr;

  for (tnum = 0; tnum < ntemplates; tnum++) {
    t_ptr = &(templates[tnum]);
    if (input_ptr->nframes - t_ptr->nframes < 0) {
      fprintf(stderr, "match: skipping template `%s' since length (%d) > input length (%d)\n",
	      t_ptr->info_string, t_ptr->nframes, input_ptr->nframes); }
    else {
      /* done for side-effect: sets best_match_yet, helps effective pruning */
      if (most_likely_offset + t_ptr->nframes < input_ptr->nframes)
	match_at_time(input_ptr, t_ptr, most_likely_offset); 

      for(index = 0; index + t_ptr->nframes <= input_ptr->nframes; index++) {
	  match_at_time(input_ptr, t_ptr, index);} 
    } } }


/* ============================================================================= */
/* ========= some stuff called only from real.c ========== */

static int next_frame_to_match;

/* ----------------------------------------------------------------------------- */
void init_match(struct signature *s_ptr)
{
  if (realtime_winp)  prepare_realtime_canvas(ntemplates, templates);
  usleep(50000);     /* this shouldn't be necessary */
  /* see if the header-type stuff of the input
     (as determined by the command-line arguments, and stored in input_sig )
     is the same as that of the templates */
  check_compatibility(s_ptr);
  pre_match_initialization();
  next_frame_to_match = 0;
}

/* ----------------------------------------------------------------------------- */
void match_some(int last_valid_frame, struct signature *input_ptr)
{
  if (verbosep) fprintf(stderr, "  matching from nftm %d thru lvf %d\n", 
	  next_frame_to_match, last_valid_frame); 
  incremental_match(next_frame_to_match, last_valid_frame, input_ptr); 
  next_frame_to_match = last_valid_frame + 1;
}

/* ----------------------------------------------------------------------------- */
void finish_match_display(struct signature *s_ptr, FILE *output_fp)
{
  s_ptr->nframes = next_frame_to_match - 1;
  if (verbosep) fprintf(stderr," mmeat: processed %d frames\n", s_ptr->nframes);
  if (realtime_winp) {
    draw_y_axis(2, ntemplates);        /* should be redundant, but is required */
    draw_template_labels(2, ntemplates, templates);
    draw_right_template_labels(ntemplates, templates, s_ptr->nframes);
    draw_right_y_axis(ntemplates, s_ptr->nframes); 
    sg_flush(); }
}

/* ----------------------------------------------------------------------------- */
void match_cleanup(void)
{
  if (realtime_winp) { 
    fprintf(stderr, "  all done ... just pausing 10 sec so you can see the window\n");
    sleep(10); 
    clean_up_realtime_canvas(); } }



/* ============================================================================= */
/* this code assumes that silence only occurs at begining and end of the file */

/* the window width can't exceed this because it must fit within some true silence,
   and the silence grab puts in a file is only (conservatively) 100 ms */
#define SILENCE_WINDOW_WIDTH  10     /* frames */
#define SW_HALF_WIDTH  (SILENCE_WINDOW_WIDTH / 2)
/* amounts by which energy can exceed lowest level and still be considered silence. 
   these numbers are probably very specific to my exact recording environment */
#define STRICT_SILENCE_CRITERION .25  
#define LOOSE_SILENCE_CRITERION .45  
#define STRICT_SILENCE_HYP_SCORE .5
#define LOOSE_SILENCE_HYP_SCORE .1

void detect_silence(struct signature *i_ptr)
{
  int center, i;
  float sum[MAX_FRAMES];      /* weighted average energy in window */
  float sorted_sum[MAX_FRAMES];      
  float typical_high, typical_low;
  float strict_threshold, loose_threshold;  /* silence thresholds */
  int last_non_silence_frame;
  int nframes = i_ptr->nframes;

  if (!i_ptr->logp) {
    fprintf(stderr, "sorry, can't form silence hypotheses unless feat file was made with log energy");
    exit(STRANGE); }
  
  for (center = SW_HALF_WIDTH; center < nframes - SW_HALF_WIDTH; center++) {
    sum[center] = 0.0;
    for (i = - SW_HALF_WIDTH; i < SW_HALF_WIDTH; i++) {
      sum[center] += 
	(1 - ( abs(i) / SW_HALF_WIDTH) )  /* gaussian curve might be nicer */
	  * i_ptr->energy[center + i]; }
    /* printf("sum[%d] is\t %.2f,\n  ", center, sum[center]); */
    sorted_sum[center] = sum[center];
  }

  qsort(sorted_sum + SW_HALF_WIDTH, nframes - 2 * SW_HALF_WIDTH, sizeof(float), 
	fl_bigger);

  typical_low = (sorted_sum[SW_HALF_WIDTH + 0] + sorted_sum[SW_HALF_WIDTH + 1] 
    + sorted_sum[SW_HALF_WIDTH + 2]) / 3.0;
  typical_high = (sorted_sum[nframes - SW_HALF_WIDTH - 1]
		  + sorted_sum[nframes - SW_HALF_WIDTH - 2] 
		  + sorted_sum[nframes - SW_HALF_WIDTH - 3]) / 3.0;
  /* printf(" typical_low: %.2f, typical_high: %.2f\n", typical_low, typical_high);*/

  strict_threshold = 
    (1 - STRICT_SILENCE_CRITERION) * typical_low + 
      STRICT_SILENCE_CRITERION * typical_high;
  loose_threshold = 
    (1 - LOOSE_SILENCE_CRITERION) * typical_low + 
      LOOSE_SILENCE_CRITERION * typical_high;
  
  /* printf(" strict_threshold is %.2f, loose_threshold is %.2f, \n", 
	 strict_threshold, loose_threshold); */

  /* a little display of where silence is detected */
  /*  for (center = SW_HALF_WIDTH; center < nframes - SW_HALF_WIDTH; center++) {
    if (center % 10 == 0)       printf(" (fr%3d) ", center);
    if (sum[center] < strict_threshold)     printf(".");  
    else if (sum[center] < loose_threshold)     printf(":");  
    else       printf("!"); } */

  make_template("#-detected-by-energy", 0, 
		detect_first_non_silence_frame(strict_threshold, sum, nframes),
		STRICT_SILENCE_HYP_SCORE);
  make_template("#-detected-by-energy", 0, 
		detect_first_non_silence_frame(loose_threshold, sum, nframes),
		LOOSE_SILENCE_HYP_SCORE);

  last_non_silence_frame = 
    detect_last_non_silence_frame(strict_threshold, sum, nframes);
  make_template("#-detected-by-energy", last_non_silence_frame + 1, 
		nframes - last_non_silence_frame, STRICT_SILENCE_HYP_SCORE);
  last_non_silence_frame = 
    detect_last_non_silence_frame(loose_threshold, sum, nframes);
  make_template("#-detected-by-energy", last_non_silence_frame + 1, 
		nframes - last_non_silence_frame, 
		LOOSE_SILENCE_HYP_SCORE);
}

/* ----------------------------------------------------------------------------- */
int detect_first_non_silence_frame(float threshold, float sum[], int nframes)
{
  int frame = SW_HALF_WIDTH;
  while (sum[frame] < threshold && frame < nframes) {
    frame++; }
  return(frame); }

int detect_last_non_silence_frame(float threshold, float sum[], int nframes)
{
  int frame = nframes - SW_HALF_WIDTH;
  while (sum[frame] < threshold && frame > 0) {
    frame--; }
  return(frame); }

/* ----------------------------------------------------------------------------- */
void make_template(char *info_string, int start_frame, int nframes, float score)
{
  strcpy(templates[nwh].info_string, info_string);
  templates[nwh].best_offset = start_frame;
  templates[nwh].nframes = nframes;
  templates[nwh].best_score = score;
  nwh ++;
}

/*=============================================================================*/
