/* file = showmatch.c */ 
/* graphically shows the word spotting results.
   reads in a .wh file, and the corresponding .la and .au files, if any */ 
/* Nigel Ward, University of Tokyo, April 1994 */

/*-----------------------------------------------------------------------------*/
#include "jmeat.h"  /* this in turn includes cheap.h and util.h */
#include "xutil.h"
#include "auutil.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

/*----------------------------------------------------------------------------- */
#define CANVAS_WIDTH      600
#define CANVAS_HEIGHT     900

#define FILENAME_XPOS    10
#define LEFT_MARGIN      20

#define FILENAME_YPOS     15
#define INPUT_LEN_YPOS  (FILENAME_YPOS + 17)
#define GRAPH_BASE (INPUT_LEN_YPOS + 25)
#define LABEL_YPOS (GRAPH_BASE + 30)
#define YSPACING     14
#define RESULT_XPOS 300
#define LABEL_STRING_OFFSET 6

#define GRAPH_SCALE 20

#define NOT_FOUND -1

/*----------------------------------------------------------------------------- */
extern Display	*display;
extern Window	win;
extern XEvent  report;

/* extern struct signature templates[MAX_NTEMPLATES]; */

/*-----------------------------------------------------------------------------*/
void showmatch_usage(void); 
int process_showmatch_options(int argc, char *argv[]);
void main(int argc,char *argv[]);
void event_loop(void);
void handle_command(char *string);          
void print_help(void);
void clean_up(void);
int ms_to_pixels(int milliseconds);   
void redraw_display(void);
void display_labels(void);
void display_hyps(void);
void display_filename(void);
void draw_notice(char *string);
void draw_label(int where, char *string);   
void play_audio(void);
void open_next_file(void);
void open_previous_file(void);
void do_new_file(void); 
void display_one_hyp(struct whyp *hyp_ptr, int index, int ypos);
void show_array(int ybase, float *array, int redp);

/*-----------------------------------------------------------------------------*/
static float ms_per_pixel;

static int nfiles;
static int active_file;
static char file_list[100][MAX_PATH_LEN];
/* directory in which to seek .la and .au files */
static char label_dir[MAX_PATH_LEN];   

static int right_margin; 
int verbose;
int log_display;

/*=============================================================================*/
void showmatch_usage(void) 
{
  fprintf(stderr, "usage: \n");
  fprintf(stderr, "   showmatch [-s number] [-r directory] [-v] [-g] [input_files.wh]+\n");
  fprintf(stderr, "Options \n");
  fprintf(stderr, "  -r <dir>   specify directory to use for .la and .au files\n");
  fprintf(stderr, "  -s <dir>   x scale factor (milliseconds per pixel)\n");
  fprintf(stderr, "  -v         verbose: also show timelines for each word\n"); 
  fprintf(stderr, "  -l         show log scores\n"); 
}

int process_showmatch_options(int argc, char *argv[])
{
  int i, last_arg_read;
  last_arg_read = 0;
  ms_per_pixel = DEFAULT_MS_PER_PIXEL;
  strcpy(label_dir, "");

  for (i = 1; i < argc && argv[i][0] == '-'; i++ ) {
    /* fprintf(stderr, ", argv[%d]= %s\n", i, argv[i]); */
    switch(argv[i][1]) {
    case 'r':  i++;  strcpy(label_dir, argv[i]);  break;
    case 's':  i++;  sscanf(argv[i], "%f", &ms_per_pixel);  break;
    case 'v':  verbose = TRUE; break;
    case 'l':  log_display = TRUE; break;
    default:  
      fprintf(stderr, "   (%c is not a known option)\n", argv[i][1]);
      showmatch_usage(); exit(STRANGE); break;
    }
    last_arg_read = i;
  }
  return(last_arg_read);
}


/*-----------------------------------------------------------------------------*/
void main(int argc,char *argv[])
{
  int i, first_wh_file;

  first_wh_file = 1 + process_showmatch_options(argc, argv);
  
  nfiles = argc - first_wh_file;
  /* fprintf(stderr,"argc %d, first_wh_file %d, nfiles %d\n",
	  argc, first_wh_file, nfiles); */

  if (nfiles == 0) {
    fprintf(stderr, "please specify one or more .wh files\n");
    showmatch_usage();
    exit(STRANGE);}
  for (i = 0; i < nfiles; i++) {
    strcpy(file_list[i], argv[first_wh_file + i]); }
  active_file = 0;
  sg_create(CANVAS_WIDTH, CANVAS_HEIGHT, 0, "showmatch");
  print_help();
  do_new_file();
  sg_flush();     /* sometimes the window comes up blank; this doesn't help */
  event_loop(); 
  /* not reached */
  exit(NORMAL);
}

/*-----------------------------------------------------------------------------*/
void event_loop(void)
{
  KeySym keysym;
  static char buffer[MAX_WORD_LEN];
  XComposeStatus compose;
  char *string;

  /* event loop code taken from Nye, volume 2 pg 72 */
  XSelectInput(display, win,  ExposureMask | KeyPressMask | ButtonPressMask | StructureNotifyMask);

  while (TRUE) {
    XNextEvent(display, &report);
    switch (report.type) {
    case Expose:
	if (report.xexpose.count != 0)
	  break;
	redraw_display();
	break;
    case KeyPress:
	/* it's a command */
	(void) XLookupString( (XKeyEvent *) &report, buffer, MAX_WORD_LEN, &keysym, &compose); 
	string = XKeysymToString(keysym);
	if (string != NULL)
	  handle_command(string);
	break; 
    }
  }
}

/*-----------------------------------------------------------------------------*/
void handle_command(char *string)          
{
  char cmd;
  char msg_string[100];

 if (strlen(string) == 1)     
    cmd = string[0];
  else {
    if (strcmp (string, "space") == 0) cmd = ' ';
    else if (strcmp (string, "semicolon") == 0) cmd = ';';
    else if (strcmp (string, "comma") == 0) cmd = ',';
    else if (strcmp (string, "BackSpace") == 0) cmd = '9';
    else if (strcmp (string, "Delete") == 0) cmd = '9';
    else {    
      sprintf(msg_string, "unknown command `%s': see term window for command list", string);
      draw_notice(msg_string);
      print_help();
      return ; } }

  switch (cmd) {
  case 'q': clean_up(); exit(NORMAL); break;
  case 'o': case 'n': case ' ': open_next_file(); break;
  case 'u': case 'p': case '9': open_previous_file(); break;
  case 'a': play_audio();    break;
  default: 
    sprintf(msg_string, "unknown command: %c; see term window for commands", cmd);
    draw_notice(msg_string);
    redraw_display();
    print_help();
    break;  
} }

/*-----------------------------------------------------------------------------*/
/* should write to window, not the kterm */
void print_help(void)
{
  fprintf(stderr, "Showmatch Commands:  (to type at the window)  \n");
  fprintf(stderr, "    r  -  redraw display                    \n");
  fprintf(stderr, "    o,n,space  -  open next file            \n");
  fprintf(stderr, "    u,p,backspace  -  reopen previous file  \n");
  fprintf(stderr, "    a  -  play audio (to speaker or jack) \n");
  fprintf(stderr, "    q  -  quit                            \n"); }
  

/*=============================================================================*/

void clean_up(void)
{ sg_close();}

/*-----------------------------------------------------------------------------*/
int ms_to_pixels(int milliseconds)   
{  return( (int) (LEFT_MARGIN + (milliseconds / ms_per_pixel))); }

/*-----------------------------------------------------------------------------*/
void redraw_display(void)
{
  sg_clear();
  display_filename();
  display_labels();
  display_hyps(); }

/*-----------------------------------------------------------------------------*/
void display_labels(void)
{
  int i, y1, y2;
  y1 = GRAPH_BASE + 5;
  y2 = y1 + 50 + YSPACING * nwh;

  vertical_dots(ms_to_pixels(0), y1, y2);
  for (i = 0; i < nlabels; i++) {
    draw_label(LABEL_STRING_OFFSET + ms_to_pixels(labels[i].start), labels[i].label);
    vertical_dots(ms_to_pixels(labels[i].start), y1, y2); 
    vertical_dots(ms_to_pixels(labels[i].end), y1, y2); 
  }
  vertical_dots(ms_to_pixels(input_length), y1, y2);
 }

/*-----------------------------------------------------------------------------*/
void display_hyps(void)
{
  int i, t, xpos, ypos, zeroth_whyp_ypos;

  right_margin = ms_to_pixels(input_length);

  if (nlabels != 0) {
    sg_string(RESULT_XPOS, INPUT_LEN_YPOS, score_string()); 
    for (t = 0; t < input_length; t += JUDGE_INTERVAL) {
      xpos = ms_to_pixels(t);  }
    show_array(GRAPH_BASE, correctness, TRUE);
  }
  if(verbose) {
    for (i = 0; i < nvocab; i++) {
      ypos = LABEL_YPOS + 22 + i * GRAPH_SCALE;
      show_array(ypos, wordline[i].evidence, FALSE);
      sg_string(right_margin + 10, ypos, wordline[i].word); }
    zeroth_whyp_ypos = LABEL_YPOS + 22 + nvocab * GRAPH_SCALE; }
  else
    zeroth_whyp_ypos = LABEL_YPOS + 30;

  for(i = 0; i < nwh; i++) {
    display_one_hyp(&hyps[i], i, zeroth_whyp_ypos + YSPACING * i); }
}

/*-----------------------------------------------------------------------------*/
void show_array(int ybase, float *array, int redp)
{
  int t, xpos;
  for (t = 0; t < njudgepoints; t ++){
    xpos = ms_to_pixels(t * JUDGE_INTERVAL);
    if (redp)
      sg_line_red(xpos, ybase, xpos, ybase - (int) (GRAPH_SCALE * array[t]));
    else
      sg_line(xpos, ybase, xpos, ybase - (int) (GRAPH_SCALE * array[t])); } }

    
/*-----------------------------------------------------------------------------*/
void display_one_hyp(struct whyp *hyp_ptr, int index, int ypos)
{
  int i, xpos;
  char string_for_index[3];
  char tmp_string[50];

  sprintf(string_for_index, "%2d", index);
  sg_string(LEFT_MARGIN - 16, ypos, string_for_index);
  sg_string(right_margin + 20, ypos, string_for_index);
  if (index % 3 == 0)
    horizontal_dots(LEFT_MARGIN, right_margin + 300, ypos - YSPACING + 2);
  
  if (log_display) {
    /* probably should do something intelligent if score=0 */
    sprintf(tmp_string, "%s %.2f", hyp_ptr->label, log10(hyp_ptr->score)); }
  else
    sprintf(tmp_string, "%s %.3f", hyp_ptr->label, hyp_ptr->score);

  sg_string(ms_to_pixels(hyp_ptr->start), ypos, tmp_string);
  sg_line(ms_to_pixels(hyp_ptr->start), ypos + 2, ms_to_pixels(hyp_ptr->end), ypos + 2);

  /* underline region of hypothesis which is correct */
  for (i = first_judgepoint_after(hyp_ptr->start); 
       i < first_judgepoint_after(hyp_ptr->end); 
       i++) {
    if (strcmp(label_at_time(i), hyp_ptr->label) == 0) {
      xpos = ms_to_pixels(i * JUDGE_INTERVAL);
      sg_line_red(xpos, ypos + 4, xpos + 1, ypos + 4); } }
}
  
/*-----------------------------------------------------------------------------*/
void display_filename(void)
{  char size_string[100];
   sg_string(FILENAME_XPOS, FILENAME_YPOS, file_list[active_file]);
   sprintf(size_string,"  input length = %d, %d word-hypotheses", input_length, nwh);
   sg_string(LEFT_MARGIN, INPUT_LEN_YPOS, size_string);
}

/*-----------------------------------------------------------------------------*/
/* show warning messages etc. */
void draw_notice(char *string)
{  
  sg_erase_rect(300, 0, 1000, 20);    /* blank out previous messages */
  sg_string_red(300, 15, string);}

/*-----------------------------------------------------------------------------*/
void draw_label(int where, char *string)   
{  sg_string(where, LABEL_YPOS, string);}

/*-----------------------------------------------------------------------------*/
void play_audio(void)
{
  char raw_data[MAX_SAMPLES];
  char filename[50], au_filename[50];
  int nsamples;
  char info_string[MAX_INFO_STRING_LEN];

  if (strcmp(label_dir,"") == 0) {
    strcpy(filename, file_list[active_file]); }
  else {
    strcpy(filename, label_dir);
    strcat(filename, "/");
    strcat(filename, file_list[active_file]); }
  make_new_path(au_filename, filename, ".au");

  nsamples = read_au_into_array(au_filename, info_string, raw_data, MAX_SAMPLES);
  play_audio_data(raw_data, 0, nsamples); 
}

/*-----------------------------------------------------------------------------*/
void open_next_file(void)
{
  if (active_file + 1 >= nfiles) {
    draw_notice("no more files; q to exit"); }
  else {
    active_file +=1;
    do_new_file(); } }

/*-----------------------------------------------------------------------------*/
void open_previous_file(void)
{
  if (active_file == 0) {
    draw_notice("this is the first file, no previous ones"); }
  else {
    active_file -=1;
    do_new_file(); } }

/*-----------------------------------------------------------------------------*/
void do_new_file(void) {
  int tmp1, tmp2;

  tmp1 = read_la_file_from_dir(label_dir, file_list[active_file], &nlabels, labels, TRUE);
  tmp2 =  read_wh_file(file_list[active_file], &nwh, hyps);
  if (tmp1 > 0)
    input_length = tmp1;
  else
    input_length = tmp2;
  if (nwh == 0) {
    fprintf(stderr, "*** No word hypotheses in file"); }
  setup_wordlines();
  do_judge();
  redraw_display(); }

/*-----------------------------------------------------------------------------*/
/* given a word, looks it up in the veridical list (labels)
   and return the index. If it doesn't occur there, so state.
   Assume each word appears only once per sentence. */
int lookup_correct(char *word, struct whyp labels[MAX_LABELS], int nlabels)
{
  int i;
   printf("looking up `%s', \n ", word); 

  for(i = 0; i < nlabels; i++) {
     printf(" labels[%d]=`%s';  ", i, labels[i].label);   
    if (strcmp(word, labels[i].label) == 0)
      return (i); }
  return(NOT_FOUND);
}
/*-----------------------------------------------------------------------------*/
