/* file = util.c */
/* Nigel Ward, University of Tokyo, April 1994 */ 
/* contents:
   general file i/o
   i/o for .au files 
   i/o for .fe files 
   i/o for .la files 
   i/o for .wh files 
   */

/* ----------------------------------------------------------------------------- */
#include "cheap.h"
#include "util.h"
#include <sys/file.h> 
#include <sys/stat.h>

#define EMPTY -2

/* ----------------------------------------------------------------------------- */

int parse_line(char *line, struct concept *c_ptr);
void init_tokenizer(char *line);
char *next_token(void);
int is_whitespace_char(char character);
int valid_abbrev(char *abbrev);

/* ----------------------------------------------------------------------------- */


static char fe_id_string[] = "<Speech_data_in_Feb95_fe_format>";
static char wh_id_string[] = "<Speech_data_in_Feb95_wh_format>";
static char la_id_string[] = "<Speech_data_in_Feb95_la_format>";
static char cv_id_string[] = "<Meaning_vector_file_in_cv_format>";

/*=============================================================================*/

/* allows us to create a new_path like lola.fe
   given path = lola.au, lola.wh, lola.fe, or lola, and extension = ".fe" 
   Note that extensions must have 3 characters (including '.') */
void make_new_path(char *new_path, char *path, char *extension)
{
  int i;
  strip_off_extension(path, new_path);
  for (i = 0; new_path[i] != NULL; i++);
  new_path[i  ] = extension[0];
  new_path[i+1] = extension[1];
  new_path[i+2] = extension[2];
  new_path[i+3] = NULL; }

/*-----------------------------------------------------------------------------*/
void strip_off_extension(char *filename, char *new)
{
  int i;
  strcpy(new, filename);
  for (i = strlen(new); i > 0; i--) {
    if (new[i] == '.') {
      new[i] = NULL;
      break; } } }

/*-----------------------------------------------------------------------------*/
/* example: converts /home/sanpon/nigel/speech/test/lola.au to lola */
void basefilename(char *filename, char *pathstring)
{
  char tmpstring[100];
  int i, last_slash_pos = 0;
  
  strip_off_extension(pathstring, tmpstring);
  
  /* strip off all directories, if any */
  for (i = 0 ; i < strlen(tmpstring); i++)
    if (tmpstring[i] == '/')
      last_slash_pos = i;
  if (last_slash_pos != 0) {
    for (i = 0; tmpstring[i] != NULL; i++)
      filename[i] = tmpstring[i + last_slash_pos + 1]; }
  else strcpy(filename,tmpstring);
}

/* ----------------------------------------------------------------------------- */
extern char *sys_errlist[];
/* ----------------------------------------------------------------------------- */
FILE *fopen_with_check(char *file_name, char *mode_string)
{ FILE *in_fp;
  in_fp = fopen(file_name, mode_string);
  if ( (int) in_fp == NULL) {
    perror(file_name);
    return(0); }
  return(in_fp); }

/* ----------------------------------------------------------------------------- */
FILE *fopen_silently(char *file_name, char *mode_string)      
{ FILE *in_fp;
  in_fp = fopen(file_name, mode_string);
  if ( (int) in_fp == NULL) {
    return(0); }
  return(in_fp); }


/* ----------------------------------------------------------------------------- */
FILE *fopen_or_exit(char *file_name, char *mode_string)
{ FILE *in_fp;
  in_fp = fopen(file_name, mode_string);
  if ( (int) in_fp == NULL) {
    perror("allegedly");
    fprintf(stderr, "Can't open file `%s' ... exiting \n", file_name);
    exit(STRANGE); }
  return(in_fp); }


/*-----------------------------------------------------------------------------*/
int open_rdonly_with_check(char *filename)
{
  int fd;
  fd = open(filename, O_RDONLY);
  if (fd < 0) {
    printf("sorry, can't open file `%s',\n", filename); 
    perror(" "); }
  return(fd); }

/*-----------------------------------------------------------------------------*/
FILE *fopen_out_file(char *path, char *extension)      
{
  char new_path[MAX_PATH_LEN]; 
  make_new_path(new_path, path, extension);
  return(fopen_with_check(new_path, "w")); }

/* ----------------------------------------------------------------------------- */
int good_file_format(FILE *in_fp, char *id_string, char *filename)   
{
  char misc_string[100], file_id_string[100];
  fgets(misc_string, sizeof(misc_string), in_fp);
  sscanf(misc_string, "%s", file_id_string);
  if (strcmp(file_id_string, id_string) != 0) {
    fprintf(stderr, "   expected      `%s' \n", id_string);
    fprintf(stderr, "   but file had  `%s' \n", file_id_string);
    fprintf(stderr, "   (file=`%s')\n", filename);
    fclose(in_fp);
    return(0); }
  return(1); }

/* ----------------------------------------------------------------------------- */
int open_new_file(char *filename)
{
  return(open(filename, O_WRONLY | O_CREAT | O_TRUNC, 0644)); 
}


/*=============================================================================*/
/* write header for .fe (in format expected by match.c) */

void write_fe_header(FILE* fp, struct signature *sig_ptr)
{
  fprintf(fp, "%s\n", fe_id_string);
  fprintf(fp, "number_of_samples: %d \n", sig_ptr->nsamples);
  fprintf(fp, "number_of_frames: %d \n", sig_ptr->nframes);
  if(sig_ptr->info_string[0] == NULL)
    fprintf(fp, "template_name: %s \n", "unknown");
  else
    fprintf(fp, "template_name: %s \n", sig_ptr->info_string);
  fprintf(fp, "features_per_frame: %d \n", sig_ptr->nbins);
  fprintf(fp, "features_from: %s \n", sig_ptr->feature_type);
  fprintf(fp, "frame_period: %d \n", sig_ptr->frame_period);
  fprintf(fp, "samples_per_frame: %d \n", sig_ptr->samples_per_frame);
  fprintf(fp, "samples_per_ms: %d \n", sig_ptr->samples_per_ms);
  fprintf(fp, "log: %d \n", sig_ptr->logp);
  fprintf(fp, "data:\n");
}

/* ----------------------------------------------------------------------------- */
/* write the featurized version of one frame of speech to the output file */
void write_fe_data(FILE *out_fp, struct signature *s_ptr)
{
  int frame, bin;

  for (frame = 0; frame < s_ptr->nframes; frame++) {
    write_fe_frame_number(out_fp, frame);
    for (bin = 0; bin < s_ptr->nbins; bin++) {
      write_fe_bin(out_fp, s_ptr->bins[frame][bin]); }
    write_fe_energy(out_fp, s_ptr->energy[frame]); }
  fflush(out_fp);     /* not strictly necessary */
}

void write_fe_frame_number(FILE *out_fp, int frame)
{     fprintf(out_fp, "frame %3d  ", frame); }

void write_fe_bin(FILE *out_fp, float value)
{     fprintf(out_fp, " %7.4f ", value); }

void write_fe_energy(FILE *out_fp, float value)
{     fprintf(out_fp, " ttl: %8.4f\n", value);}


/* ----------------------------------------------------------------------------- */
int read_fe_into_sig(FILE *fp, struct signature *tptr)
{
  int frame, bin, status, exit_flag1, exit_flag2, flag;
  float value, energy;
  char misc_string[100], keyword[50], info[50];

  if (!good_file_format(fp, fe_id_string, "")) {
    fprintf(stderr,"        apparently not a fe; skipping it. \n "); 
    return(FALSE);} 

  /* grab the information from the header.  Don't check for missing fields,
     since feat etc are trustworthy */
  exit_flag1 = FALSE;
  while (!exit_flag1) {
    fgets(misc_string, sizeof(misc_string), fp);
    sscanf(misc_string, " %s %s", keyword, info);
    if (strcmp(keyword, "number_of_samples:") == 0)
      { tptr->nsamples = atoi(info); }
    else if (strcmp(keyword, "samples_per_frame:") == 0)
      { tptr->samples_per_frame = atoi(info); }
    else if (strcmp(keyword, "samples_per_ms:") == 0)
      { tptr->samples_per_ms = atoi(info); }
    else if (strcmp(keyword, "number_of_frames:") == 0)
      { /* ignore it, since we get frame count from actually reading */ }
    else if (strcmp(keyword, "features_from:") == 0) {
      strcpy(tptr->feature_type,info);  }
    else if (strcmp(keyword, "features_per_frame:") == 0) {
      tptr->nbins = atoi(info);  }
    else if (strcmp(keyword, "template_name:") == 0) {
      strcpy(tptr->info_string, info);  }
    else if (strcmp(keyword, "frame_period:") == 0) {
      tptr->frame_period = atoi(info);   }
    else if (strcmp(keyword, "log:") == 0) {
      tptr->logp = atoi(info);   }
    else if (strcmp(keyword, "data:") == 0) {
      /* stop processing header and start reading data */
      exit_flag1 = TRUE;  }
    else {fprintf(stderr,"match: ignoring unknown fe header field: `%s'\n", keyword);}
  }
  /* fprintf(stderr, "    data_length  = %2d,  tptr->info_string = `%s', \n",
      data_length, tptr->info_string);
     fprintf(stderr, "    tptr->nbins = %d,  tptr->feature_type = `%s' \n", 
      tptr->nbins, tptr->feature_type);  */

  exit_flag2 = FALSE;
  /* read in the frames, each of form ``frame: XX   XX XX XX XX  ttl: XX'' */
  for (frame = 0; exit_flag2 == FALSE && frame < MAX_FRAMES; frame++) {
    status = fscanf(fp, " frame %d ", &flag);
    if (status == EOF) {
      /* fprintf(stderr,"    match: normal EOF after frame %d\n", frame); */ 
      tptr->nframes = frame;
      return(TRUE); }
    else if (status == 0) 
      {fprintf(stderr,"match: failed to find frame start. \n"); return(FALSE);}
    else {
      for (bin = 0; bin < tptr->nbins; bin++) {
	status = fscanf(fp, " %f ", &value);
	if (status == EOF  || status == 0) 
	  {fprintf(stderr,"match: !! fscanf returned %d (frame %d, bin %d) \n",
		   status, frame, bin);
	   return(FALSE);}
	else
	  tptr->bins[frame][bin] = value; } }
    fscanf(fp, " ttl: %f ", &energy);
    tptr->energy[frame] = energy; 
  }
  /* not reached, statement present only to satisfy lint */
  return(TRUE);
}

/*=============================================================================*/
void write_la_file_header(FILE *out_fp, char *input_info_string, int input_length)
{  fprintf(out_fp, "%s\n", la_id_string);
   fprintf(out_fp, "input_info_string: %s\n", input_info_string);
   fprintf(out_fp, "input_length: %d \n", input_length);
   fprintf(out_fp, "data:\n");
 }

void write_la_entry(FILE *out_fp, char *label, int start, int end)
{  fprintf(out_fp, "%s %d %d\n", label, start, end); }


FILE *la_open_file_check_header(char *sister_filename, int silentp)
{
  FILE *in_fp;
  char la_file_name[MAX_PATH_LEN];

  make_new_path(la_file_name, sister_filename, ".la");
  if (silentp)
    {in_fp = fopen_silently(la_file_name, "r");}
  else
    {in_fp = fopen_with_check(la_file_name, "r");}
  if ((int) in_fp == NULL) return (0);
  else if (good_file_format(in_fp, la_id_string, la_file_name))
    return (in_fp);
  else return(0); }


/*-----------------------------------------------------------------------------*/
int read_la_file(char *filename, int *nlabels_ptr, struct whyp labels[], int silentp)
{
  FILE *label_fp;   
  int nlabels, input_length;
  int status, exit_flag1, exit_flag2;
  char misc_string[200], keyword[50], info[50];

  label_fp = (FILE *) la_open_file_check_header(filename, silentp);
  if (label_fp == 0) {return(0);}

  
  exit_flag1 = FALSE;
  while (!exit_flag1) {
    fgets(misc_string, sizeof(misc_string), label_fp);
    sscanf(misc_string, " %s %s", keyword, info);
    /* fprintf(stderr, "read in: `%s'  and  `%s'\n", keyword, info);  */
    if (strcmp(keyword, "input_length:") == 0)
      {input_length = atoi(info);
       if (input_length == 0) {
	 fprintf(stderr, "Sorry, input_length is not specified in `%s', exiting\n", 
		 filename);
	 exit(STRANGE); } }
    else if (strcmp(keyword, "input_info_string:") == 0)
      {/* ignore input info string */}
    else if (strcmp(keyword, "data:") == 0) {
      /* stop processing header and start reading data */
      exit_flag1 = TRUE; }
    else {fprintf(stderr,"*** ignoring unknown la header field: `%s'\n", keyword);}
  }

  nlabels = 0;
  exit_flag2 = FALSE;
  while(!exit_flag2) {
    status = fscanf(label_fp, "%s %d %d ",  labels[nlabels].label, 
		    &labels[nlabels].start, &labels[nlabels].end);
    if (status == EOF)
      {exit_flag2 = TRUE;}
    else if (status == 0)
      {fprintf(stderr, " *** read failed\n"); exit_flag2 = TRUE;}
    else {
      labels[nlabels].len = labels[nlabels].end - labels[nlabels].start;
      nlabels++; }
  }
  
  *nlabels_ptr = nlabels;
  fclose(label_fp);
  return(input_length);
}


int read_la_file_from_dir(char *dirname, char *filename, int *nlabels_ptr,
			  struct whyp labels[], int silentp)
{
  char new_filename[MAX_PATH_LEN], last_component_only[MAX_PATH_LEN];
  if (strcmp(dirname, "") == 0) 
    return(read_la_file(filename, nlabels_ptr, labels, silentp));
  else {
    basefilename(last_component_only, filename);
    strcpy(new_filename, dirname);
    strcat(new_filename, "/");
    strcat(new_filename, last_component_only);
    strcat(new_filename, ".la");
    return(read_la_file(new_filename, nlabels_ptr, labels, silentp)); }
}

/*=============================================================================*/
int read_wh_file(char *filename, int *nwh_ptr, struct whyp hyps[MAX_NTEMPLATES])
{
  int nwh, input_length;
  int status, exit_flag1, exit_flag2;
  char misc_string[200], raw_label[MAX_WORD_LEN], keyword[50], info[50];
  char file_id_string[100];
  char wh_filename[MAX_PATH_LEN];
  char *fgets_result;
  FILE *wh_fp;

  /* since now this code must handle both .wh and .gl files, we can't automatically
     coerce to a .wh extension, so we don't coerce at all */ 
  /* make_new_path(wh_filename, filename, ".wh"); */
  strcpy(wh_filename, filename);

  wh_fp = fopen_with_check(wh_filename, "r");
  if (wh_fp == 0)
    {fprintf(stderr, "couldn't open wh file `%s'", wh_filename); 
     perror(" ");
     exit(STRANGE);}

  fgets_result = fgets(misc_string, sizeof(misc_string), wh_fp);
  if (fgets_result == NULL) {
    fprintf(stderr, " empty file! (`%s') \n", filename);
    exit(STRANGE); }
  sscanf(misc_string, "%s", file_id_string);
  if (strcmp(file_id_string, wh_id_string) != 0) {
    fprintf(stderr, " expected `%s' \n", wh_id_string);
    fprintf(stderr, " but file had  `%s' (file=`%s')\n", file_id_string, filename);
    fclose(wh_fp);
    exit(STRANGE);}
  
  exit_flag1 = FALSE;
  while (!exit_flag1) {
    fgets(misc_string, sizeof(misc_string), wh_fp);
    sscanf(misc_string, " %s %s", keyword, info);
    /* fprintf(stderr, "read in: `%s'  and  `%s'\n", keyword, info); */
    if (strcmp(keyword, "input_length:") == 0)
      {input_length = atoi(info);
       if (input_length == 0) {
	 fprintf(stderr, "Sorry, input_length is not specified in `%s', exiting\n", 
		 filename);
	 exit(STRANGE); } }
    else if (strcmp(keyword, "input_info_string:") == 0)
      {/* ignore input info string */}
    else if (strcmp(keyword, "data:") == 0) {
      /* stop processing header and start reading data */
      exit_flag1 = TRUE; }
    else {fprintf(stderr,"*** ignoring unknown wh header field: `%s'\n", keyword); }
  }

  nwh = 0;
  exit_flag2 = FALSE;
  while(!exit_flag2) {
    status = fscanf(wh_fp, " %s %d %d %f ) ",
		    raw_label, &hyps[nwh].start, &hyps[nwh].end, &hyps[nwh].score);
    if (status == EOF)
      {exit_flag2 = TRUE;}
    else if (status != 4) {
      fprintf(stderr, "fscanf returned %d; wrong number of fields in line\n",status);
      fprintf(stderr, "%d templates read in \n", nwh);
      exit_flag2 = TRUE;}
    else {
      clip_word_name(hyps[nwh].label,raw_label);
      hyps[nwh].len = hyps[nwh].end - hyps[nwh].start;

      /* printf("hypo `%s' from %d to %d with score %f\n",
	     hyps[nwh].label, hyps[nwh].start, hyps[nwh].end, hyps[nwh].score);  */
      nwh++;
      if (nwh > MAX_NTEMPLATES) {
	fprintf(stderr, "WARNING: too many hypotheses, only reading first %d\n",MAX_NTEMPLATES);
	nwh = MAX_NTEMPLATES;
	break; }
    }
  }
  fclose(wh_fp);
  *nwh_ptr = nwh;
  if (input_length == 0)     return(0);
  else    return(input_length);
}

/* extracts "john" from "john-fm-jms", "john-xxx-yyy" etc */ 
void clip_word_name(char *output, char *raw_label)
{
  int i;  char this_char;

  for (i = 0; i < MAX_WORD_LEN; i++) {
    this_char = raw_label[i];
    if (this_char  == '-' || this_char == NULL || this_char == ' ') {
      output[i] = NULL;
      break; }
    else
      output[i] = this_char; } }

/*-----------------------------------------------------------------------------*/
void write_wh_file_hdr(FILE *out_fp, char *input_info_string, int input_length)
{
  fprintf(out_fp, "%s\n", wh_id_string); 
  fprintf(out_fp, "input_info_string: %s\n", input_info_string);
  fprintf(out_fp, "input_length: %d \n", input_length);
  fprintf(out_fp, "data:\n");
}


/*-----------------------------------------------------------------------------*/
void write_wh_file_line(FILE *out_fp, char *string, int start, int length, float score)
{
  fprintf(out_fp,  " %s  %d %d  %f \n",  string, start, start + length, score);
}


/* ============================================================================= */

/*----------------------------------------------------------------------------- */
FILE *cv_open_file_check_header(char *sister_filename, char *extension, int make_namep)
{
  FILE *in_fp;
  char file_name[MAX_PATH_LEN];

  if(make_namep) {
    make_new_path(file_name, sister_filename, extension); }
  else {
    /* would be nice to add extension if it is missing */
    strcpy(file_name, sister_filename); }
  in_fp = fopen_with_check(file_name, "r");
  if ((int) in_fp == NULL) return (0);
  else if (good_file_format(in_fp, cv_id_string, file_name))
    return (in_fp);
  else return(0); }


/*----------------------------------------------------------------------------- */
int read_cv_file(char *dirname, char *filename, struct concept cvec[],
		 char *extension, int new_name)
{
  FILE *fp;
  int count, result;
  char line[200];
  char *fgets_status;

  fp = cv_open_file_check_header(filename, extension, new_name);
  if (fp == 0) 
    return(0);

  count = 0;

  while(TRUE) {
    fgets_status = fgets(line, sizeof(line), fp);
    /* printf("read line `%s'\n", line);   */
    if (fgets_status == NULL)  /* end of file */
      break;
    if(line[0] == '*') /* end */
      break;
    if(line[0] == ';') /* comment */
      continue;
    if (count > CONCEPT_ARRAY_SIZE) {
      fprintf(stderr, "Warning: too many concepts in file; only reading first %d\n",
	      CONCEPT_ARRAY_SIZE);
	break; }

    result = parse_line(line, &(cvec[count]));
    switch (result) {
    case INVALID:  fclose(fp);  return (0); break;
    case EMPTY:    fclose(fp);  return(count); break; /* ignore rest of file */
    }
    count ++;
  }
  fclose(fp);
  return(count);
}


/*-----------------------------------------------------------------------------*/
/* god alone knows why this declaration is needed */
double atof(char *str);

int parse_line(char *line, struct concept *c_ptr)
{
  char *token;
  int pfnum;

  init_tokenizer(line);
  
  /* Parse name */
  token = next_token();
  if ((int) token == INVALID) {
    /* fprintf(stderr, "no tokens in line (blank line)\n", line); */
    return(EMPTY);   }
  strcpy(c_ptr->name, token);
  
  /* Parse presence score */
  token = next_token();
  if ((int) token == INVALID) {
    fprintf(stderr, "Warning: concept score missing: line=`%s'; ignoring rest of file\n", 
	    line);
    return(INVALID);     }
  /* if not a number, this silently sets it to 0.0; tolerable but not ideal */
  /* printf("___received token `%s'___;  ", token); */
  c_ptr->presence = (float) atof(token);
  
  /* Parse pfeatures and values */ 
    for (pfnum = 0;   ; pfnum ++) {
      if (pfnum > PFEATURE_ARRAY_SIZE)
	fprintf(stderr, "Warning: more than %d pfeatures: line=`%s'; ignoring rest\n",
		PFEATURE_ARRAY_SIZE, line);
      token = next_token();
      if ((int) token == INVALID) {
	/* we've reached the end of the line */
	c_ptr->npfeatures = pfnum;
	break; }
      strcpy(c_ptr->pf_name[pfnum], token);
      
      token = next_token();
      if ((int) token == INVALID) {
	fprintf(stderr, "Warning: no value seen for feature `%s'; ignoring it\n", 
		c_ptr->pf_name[pfnum]); 
	fprintf(stderr, "  (in line `%s')\n", line);
	c_ptr->npfeatures = pfnum;
	break; }
      c_ptr->pf_score[pfnum] = atof(token);
    }
  return(0);  /* normal exit */
}


/*----------------------------------------------------------------------------- */
/* A kludgey little tokenizing package */

static char *line_to_tokenize;

void init_tokenizer(char *line)
{  line_to_tokenize = line; }

char *next_token(void)
{
  int i, j;
  static char token[100];

  j = 0;
  /* skip leading whitespace */
  while(TRUE) {
    if (line_to_tokenize[j] == NULL) {
      return( (char *) INVALID); }
    else if (is_whitespace_char(line_to_tokenize[j])) {
      j++; }
    else break;   }

  i = 0;
  while(!is_whitespace_char(line_to_tokenize[i+j]) && line_to_tokenize[i+j] != NULL) {
    token[i] = line_to_tokenize[i + j];
    i++; }
  token[i] = NULL;    /* terminate string */
  line_to_tokenize = line_to_tokenize + i + j;
  /* printf("returning token `%s';  ", token); */
  return(token);
}

int is_whitespace_char(char character)
{ return(character ==  '\n' || 	 character ==  '\t' ||   character ==  ' '); }

/*-----------------------------------------------------------------------------*/
void verify_is_directory(char *dir)
{
  struct stat sbuf;
  if (stat(dir, &sbuf) < 0) {
    printf(" `%s' not found\n", dir);
    exit(STRANGE); } 
  if ((sbuf.st_mode & S_IFMT) != S_IFDIR) {
    printf("  `%s' is not a directory\n", dir); exit(STRANGE); }
}

int file_exists(char *file)
{
  struct stat st;
  return(stat(file, &st) == 0);
}

/*=============================================================================*/
/* eof */
